print("Loading NuVote Utilities...")
local function nvdbp(...)
	local debug = CV_FindVar("nv_debug").value
	if not debug then return end
	local oldt = {...}
	local newt = {}
	for i=1,#oldt do
		local stringified = tostring(oldt[i])
		if stringified == "" then
			stringified = "<no repr>"
		end
		newt[i] = stringified
	end
	print(table.concat(newt," | "))
end
rawset(_G,"nvdbp", nvdbp)

local nvdbpo_printeds = {}
rawset(_G,"nvdbpo",function(id,...)
	if nvdbpo_printeds[id] == true then return end
	nvdbpo_printeds[id] = true
	nvdbp(...)
end)

local function nvwarn(p,text)
	CONS_Printf(p,"\130WARNING: \128"..text)
end
rawset(_G,"nvwarn", nvwarn)

local nv_enableNV = CV_RegisterVar({
	name = "nv_enableNV",
	defaultvalue = "On",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

local nv_debug = CV_RegisterVar({
	name = "nv_debug",
	defaultvalue = "Off",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

local function test_function(params,expected,func,tolerance)
	result = func.func(table.unpack(params))
	if tolerance == nil then
		if result != expected then
			error("Test failure for function "..func.name..". Expected: "..expected.." Actual: "..result)
		end
	else
		if result < (expected - tolerance) or result > (expected + tolerance) then
			error("Test failure for function "..func.name..". Expected: "..expected.." Actual: "..result.." (Tolerance: "..tolerance..")")
		end
	end
end

--gpt generated
rawset(_G,"lstrip",function(s)
    return s:match("^%s*(.*)")
end)

-- gpt generated
local function hashString(str, givenum)
    local hash = 0
	-- print(str)
    for i = 1, #str do
        hash = ((hash << 5) - hash) + str:byte(i)
    end
	local result = hash & 0xffffffff
	if givenum == nil then
		return string.format("%08x", result)
	end
	return result
end
rawset(_G,"hashString", hashString)

-- gpt generated
local function deepCopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[deepCopy(orig_key)] = deepCopy(orig_value)
        end
        setmetatable(copy, deepCopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end

rawset(_G,"cg_deepcopy",deepCopy)

-- #######################
--         Random
-- #######################

-- Machine Random
-- I'm calling G_BuildMapName(0) which calls M_Random for RNG
-- M_Random is non sync'd.  So in a way, I'm calling C's rand(). 
-- Only problem is there's no guanrantee for even distribution. 
-- I am officially a mad man, mang

-- map names could be a-z. Only using 0-7 makes rng vulnerable
-- to be heavily skew'd by socc'd maps
-- so we map the characters too. at least they will be less affected by socs
local charmap = {}
charmap["a"]=0
charmap["b"]=1
charmap["c"]=2
charmap["d"]=3
charmap["e"]=4
charmap["f"]=5
charmap["g"]=6
charmap["h"]=7
charmap["i"]=0
charmap["j"]=1
charmap["k"]=2
charmap["l"]=3
charmap["m"]=4
charmap["n"]=5
charmap["o"]=6
charmap["p"]=7
charmap["q"]=0
charmap["r"]=1
charmap["s"]=2
charmap["t"]=3
charmap["u"]=4
-- remainder characters. Uncommenting will cause skew
--charmap["v"]=5
--charmap["w"]=6
charmap["x"]=7
charmap["y"]=0
charmap["z"]=1

charmap["8"]=5
charmap["9"]=6
charmap["0"]=0
charmap["1"]=1
charmap["2"]=2
charmap["3"]=3
charmap["4"]=4
charmap["5"]=5
charmap["6"]=6
charmap["7"]=7

-- rotate mappings so that skews caused by available maps get distributed over time
-- I'm making the dumbest excuses for this to work
local function rotate_charmap()
	for char, octal in pairs(charmap) do
		charmap[char] = ($ + 1) % 8
		-- nvdbp(char,charmap[char])
	end
end
addHook("MapLoad",rotate_charmap)

local function __gen_bmn_char()
	local bmn = G_BuildMapName(0):sub(5,5)
	-- local num = tonumber(bmn)
	-- if num != nil then return num end
	return bmn
end

local function NV_M_Random()
	local bitparts = {}
	for i=1,(32 / 3) do
		local char = __gen_bmn_char()
		-- nvdbp("bmn char: "..char)
		while charmap[char] == nil do
			char = __gen_bmn_char()
		end

		local threebit = charmap[char]
		table.insert(bitparts,threebit)
	end

	-- shift in all randomly generated bits
	local result = 0
	for i, bits in ipairs(bitparts) do
		result = $ | (bits << (3 * (i-1)))
	end

	-- rotate so that chances are always mixed up
	rotate_charmap()

	return result
end

local function NV_M_RandomRange(a ,b)
    return a + abs(NV_M_Random() % (b - a + 1))
end

local function NV_M_RandomChance(fixed_pct)
	local drawing = NV_M_RandomRange(0,FRACUNIT-1)
	local mustclear = FRACUNIT - fixed_pct
	-- nvdbp("FRACUNIT","drawing","mustclear")
	-- nvdbp(FRACUNIT,drawing,mustclear)
	return drawing >= mustclear
end

local function __test_NV_M_Random_distribution()
	local freqs = {}
	local mod = 1062

	for i=1,1000000 do
		local cur = NV_M_RandomRange(1,mod)
		if freqs[cur] != nil then
			freqs[cur] = $+1
		else
			freqs[cur] = 1
		end
	end

	for res, freq in pairs(freqs) do
		local msg = res.." : "..freq
		print(msg)
	end
end

-- Plucked straight from hostmod
-- From LJ Sonik's PRNG library
local  nv_prng_t = 0
local  nv_prng_x = 6278
local  nv_prng_y = 975
-- local  nv_prng_z = 39207
local  nv_prng_w = 45678
local  nv_prng_txt = ""
local  nv_prng_txtn = 39207 -- actual value that's used in prng

local  function NV_Random()
	-- Log date 9/8/2023: Two problems so far. 1) First few calls are always the same result 2) I'm seeing patterns in map choices come up 
	-- e.g. map x always comes with map y. Maybe this implementation isn't the best for choosing maps? I will try to have the text input variable
	-- affect the immediate result. But it may introduce randomness issues, and probably doesn't solve for 2. I could also use G_BuildMapName(0)
	-- but I need to also do something special for hell, encore, and battle (?)

	nv_prng_t = nv_prng_x ^^ (nv_prng_x << 11)

	-- nv_prng_x, nv_prng_y, nv_prng_z, nv_prng_w  = nv_prng_y, nv_prng_z, nv_prng_w, nv_prng_w ^^ (nv_prng_w >> 19) ^^ nv_prng_t ^^ (nv_prng_t >> 8) -- What the fuck?


	-- nv_prng_t = nv_prng_x ^^ (nv_prng_x << 11)
	-- nv_prng_x =  nv_prng_y
	-- nv_prng_y = nv_prng_z
	-- nv_prng_z = nv_prng_w
	-- nv_prng_w =  nv_prng_w ^^ (nv_prng_w >> 19) ^^ nv_prng_t ^^ (nv_prng_t >> 8) 



	nv_prng_x, nv_prng_y, nv_prng_txtn, nv_prng_w  = nv_prng_y, nv_prng_txtn, nv_prng_w, nv_prng_w ^^ (nv_prng_w >> 19) ^^ nv_prng_t ^^ (nv_prng_t >> 8) -- What the fuck?
	return nv_prng_w
end

local function NV_RandomRange(a ,b)
    return a + abs(NV_Random() % (b - a + 1))
end
rawset(_G,"NV_RandomRange", NV_M_RandomRange)

local function update_prng_txt(text)
	-- this is technically a hash chain.
	-- cgcoin is comming up bruddahs, purchase now
	local new_input = nv_prng_txt..text
	nv_prng_txt = hashString(new_input)
	nv_prng_txtn = hashString(new_input,true)
end
rawset(_G,"update_prng_txt", update_prng_txt)

-- this might be overkill, already setting based off player times
addHook("PlayerMsg", function(source, msgtype, target, msg)
	if msgtype != 0 then return end
	update_prng_txt(msg)
end)

local function reset_prng_state()
	nv_prng_t = 0
	nv_prng_x = 6278
	nv_prng_y = 975
	-- nv_prng_z = 39207
	nv_prng_w = 45678
	nv_prng_txt = ""
	nv_prng_txtn = 0 -- actual value that's used in prng
end

local function __call_prng()
	local final = ""
	for i=1,10 do
		local num = NV_RandomRange(1,9)
		final = $..tostring(num)
	end
	return final
end

local function NV_RandomChance(fixed_pct)
	local drawing = NV_RandomRange(0,FRACUNIT-1)
	local mustclear = FRACUNIT - fixed_pct
	-- nvdbp("FRACUNIT","drawing","mustclear")
	-- nvdbp(FRACUNIT,drawing,mustclear)
	return drawing >= mustclear
end
rawset(_G,"NV_RandomChance", NV_M_RandomChance)


-- local function NV_ChooseElement(t)
-- 	local i = HM_RandomRange(1,#t)
-- 	return t[i]
-- end
-- rawset(_G,"NV_ChooseElement", NV_ChooseElement)

local function test_NV_RandomChance(pct)
	local trues = 0
	local calls = 100

	for i = 1,100 do
		if NV_RandomChance(pct * FRACUNIT / 100) then
			trues =$ + 1
		end
	end
	local result = pct..": "..trues.." / "..calls
	print(result)
end

local function NV_Random_netvars(net)
	nv_prng_t = net($)
	nv_prng_x = net($)
	nv_prng_y = net($)
	-- nv_prng_z = net($)
	nv_prng_w = net($)
	nv_prng_txt = net($)
	nv_prng_txtn = net($)
end

addHook("NetVars", NV_Random_netvars)

local function __test_text_input_influences_NVR()
	reset_prng_state()

	local vanilla = __call_prng()
	local vanilla2 = __call_prng()

	reset_prng_state()
	update_prng_txt("I am speekaing now hello")
	local text1 = __call_prng()

	update_prng_txt("Another text input")
	local text2 = __call_prng()

	nvdbp("vanilla","vanilla1","text1","text2")
	nvdbp(vanilla,vanilla2,text1,text2)
	nvdbp(vanilla == text1,vanilla2 == text2,text1 == text2,vanilla == vanilla2)

end

local function __test_NV_RandomChance(fixed_pct)
	local trues = 0
	for i=1,500 do
		NV_RandomChance()
	end
end

local function __test_NV_Random_distribution()
	local freqs = {}
	local mod = 1062

	for i=1,1000000 do
		local cur = NV_RandomRange(1,mod)
		if freqs[cur] != nil then
			freqs[cur] = $+1 -- uh oh..
		else
			freqs[cur] = 1
		end
	end

	for res, freq in pairs(freqs) do
		local msg = res.." : "..freq
		print(msg)
	end
end

local function __test_NV_Random_locality()
	-- map numbers seem to always be close to each other
	local diffs = {}
	local mod = 1062
	local prev = nil

	for i=1,1000000 do
		local cur = NV_RandomRange(1,mod)
		if prev then
			local diff = cur - prev
			if diff < 0 then
				diff = (mod - prev) + cur
			end
			if diffs[diff] != nil then
				diffs[diff] = $+1 -- uh oh..
			else
				diffs[diff] = 1
			end
		end

		prev = cur

	end

	for diff, freq in pairs(diffs) do
		local msg = diff.." : "..freq
		print(msg)
	end
end

local function __test_NV_Random_sequences()
	local ssize = 4
	local buf = {}
	local seens = {}
	local mod = (36*26+26)+100 -- = 1062

	-- 1062 choose 5 = 1.11e13, which needs 42 bits to represent. 
	-- So we really shouldn't be seeing any repeat sequences


	for i=1,1000000 do
		local nxt = NV_RandomRange(1,mod)
		table.insert(buf,nxt)
		while #buf > ssize do
			table.remove(buf,1)
		end

		if #buf < ssize then continue end
		local seq = table.concat(buf,"|")

		if seens[seq] != nil then
			seens[seq] = $+1 -- uh oh..
		else
			seens[seq] = 1
		end
	end

	local all_good = true
	for seq, freq in pairs(seens) do
		if freq > 1 then
			local msg = "We saw "..seq.." "..freq.." times"
			print(msg)
			all_good = false
		end
	end

	if all_good then print("Sequences seem good") end
	
end

local function test_NV_Random()
	-- __test_text_input_influences_NVR()
	-- test_NV_RandomChance(0)
	-- test_NV_RandomChance(100)
	-- test_NV_RandomChance(25)
	-- test_NV_RandomChance(50)
	-- test_NV_RandomChance(75)
	-- __test_NV_Random_sequences()
	-- __test_NV_Random_locality()
	-- __test_NV_Random_distribution()
	__test_NV_M_Random_distribution()
end

rawset(_G,"test_NV_Random", test_NV_Random)


-- https://stackoverflow.com/questions/5525817/inline-conditions-in-lua-a-b-yes-no
local function ternary ( cond , T , F )
    if cond then return T else return F end
end
rawset(_G,"ternary", ternary)

function string:inTable(tbl)
    for _, value in ipairs(tbl) do
        if self == value then
            return true
        end
    end
    return false
end

local function build_map_title(map)
	-- G_BuildMapTitle not available for 2.1
	local mapstr =  map.lvlttl.." "..(map.subttl and map.subtt1 or "")..(map.zonttl and map.zonttl or "")..(map.actnum and (" "..map.actnum) or "")
	mapstr = string.match(mapstr, "(.-)%s*$") -- remove trailing whitespaces
	return mapstr
end
rawset(_G,"build_map_title", build_map_title)


local function mapsfullform(maps)
	local resulttable = {}
	for i,m in ipairs(maps) do
		table.insert(resulttable,build_map_title(mapheaderinfo[m]))
	end

	return table.concat(resulttable,"|")
end
rawset(_G,"mapsfullform", mapsfullform)

-- used to split up choices that are embedded in pipe separated list
local function pipesplit(inputstr, sep)
	if sep == nil then
			sep = "%s"
	end
	local t={}
	for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
			table.insert(t, str)
	end
	return t
end
rawset(_G,"pipesplit", pipesplit)

-- https://stackoverflow.com/questions/18843610/fast-implementation-of-queues-in-lua
-- implement a queue

local List = {}
function List.new ()
  return {first = 0, last = -1}
end

function List.pushleft (list, value)
    local first = list.first - 1
    list.first = first
    list[first] = value
end
  
function List.pushright (list, value)
    local last = list.last + 1
    list.last = last
    list[last] = value
end
  
function List.popleft (list)
    local first = list.first
    if first > list.last then error("list is empty") end
    local value = list[first]
    list[first] = nil        -- to allow garbage collection
    list.first = first + 1
    return value
end
  
function List.popright (list)
    local last = list.last
    if list.first > last then error("list is empty") end
    local value = list[last]
    list[last] = nil         -- to allow garbage collection
    list.last = last - 1
    return value
end

function List.peekleft (list)
	local first = list.first
	return list[first]
end

function List.peekright (list)
	local last = list.last
	return list[last]
end

function List.size (list)
	return list.last - list.first + 1
end

-- gpt generated
function List.iterate(list)
    local i = list.first - 1
    return function()
        i = i + 1
        if i <= list.last then
            return list[i]
        end
    end
end
rawset(_G,"List", List)


-- https://stackoverflow.com/questions/2282444/how-to-check-if-a-table-contains-an-element-in-lua
local Set = {}
function Set.new()
    return {_set = {}, size=0}
end

function Set.contains(set, key)
    return set._set[key] ~= nil
end

function  Set.add(set, key)
	if Set.contains(set,key) then return end
	-- for k,v in pairs(set._set) do
	-- 	nvdbp("key:"..k)
	-- end
    set._set[key] = true
	set.size =$+1
end

function Set.remove(set, key)
	if not Set.contains(set,key) then return end
    set._set[key] = nil
	set.size =$-1
end

rawset(_G,"Set", Set)

-- maintains a queue, and will tell you if a key exists in it
-- this is a bit overengineered sorry. gpt says its good :)
local QueuedKeys = {}
function QueuedKeys.new(size)
	return {queue=List.new(),set=Set.new(),cap=size}
end

function QueuedKeys.add(qk,key)
	while qk.cap != 'inf' and List.size(qk.queue) != 0 and List.size(qk.queue) >= qk.cap do
		local expelled = List.popright(qk.queue)
		Set.remove(qk.set,expelled)
	end

	local key_ = tostring(key)
	List.pushleft(qk.queue,key_)
	Set.add(qk.set,key_)
end

function QueuedKeys.contains(qk,key)
	return Set.contains(qk.set,tostring(key))
end

function QueuedKeys.size(qk)
	return List.size(qk.queue)
end

function QueuedKeys.iterate(qk)
    return List.iterate(qk.queue)
end

function QueuedKeys.updatecapacity(qk,cap)
	if cap == 'inf' or cap >= qk.cap then return end -- nothing else needed

	while List.size(qk.queue) > cap do
		local expelled = List.popright(qk.queue)
		Set.remove(qk.set,expelled)
	end

	qk.cap = cap
end

function QueuedKeys.clear(qk)
	local temp_cap = qk.cap
	QueuedKeys.updatecapacity(qk,0)
	QueuedKeys.updatecapacity(qk,temp_cap)
end

function QueuedKeys.timetoejected(qk,key)
	-- how many other keys are ahead of the key in queue, if exists
	if not QueuedKeys.contains(qk,key) then return -1 end

	local keysahead = 0
	for i = qk.queue.first,qk.queue.last do
		if qk.queue[i] == key then return keysahead end
		keysahead = $+1
	end
end

rawset(_G,"QueuedKeys", QueuedKeys)

-- MAKE SURE TO NEVER PLAY ON EVERY TIC OR NO ONE WILL TRUST ME OR THIS MOD EVER!!!!
-- this is a safety net in case erroneous operation calls to play a sound on every tic
local prev_sounds = {}
local sound_log = {}
local banned_sounds = {}
local function NV_playsound(src_mobj,sid,dest_player)
	-- convert to pure frames. Num too large for 30 bits.
	local curr_frame = ((cg_get_current_time() % 100000) * 35) + __get_cgtime_frame()

	-- record that an attempt to play
	-- bann it if it tries to play too much within 19 frames
	if sound_log[sid] == nil then
		sound_log[sid] = List.new()
		List.pushleft(sound_log[sid],curr_frame)
	else
		local records = sound_log[sid]
		while List.size(records) > 0 and (curr_frame - List.peekright(records)) > 14 do
			List.popright(records)
		end
		List.pushleft(records,curr_frame)
		if List.size(records) > 7 then
			banned_sounds[sid] = true
		end
	end

	if banned_sounds[sid] != nil and banned_sounds[sid] then
		print("banned playing for playing a lot "..sid)
		return
	end

	-- if played less than 7 frames ago, don't play
	if prev_sounds[sid] ~= nil and (curr_frame - prev_sounds[sid]) < 5 then
		-- print("rejected playing for being within 5 fr "..sid)
		return
	end

	-- record that we played here last
	prev_sounds[sid] = curr_frame

	S_StartSound(src_mobj,sid,dest_player)
end

rawset(_G,"NV_playsound", NV_playsound)

local cgevents = {
	__events = {}
}

function cgevents.register(cge, eventname)
	if cge.__events[eventname] != nil then
		error("Event "..eventname.." already exists")
	end

	cge.__events[eventname] = {}
end

function cgevents.deregister(cge, eventname)
	cge.__events[eventname] = nil
end

function cgevents.subscribe(cge, eventname,subid,func)
	if cge.__events[eventname] == nil then
		error("Event "..eventname.." does not exist")
	end

	if eventname == nil or subid == nil or func == nil then
		error("user must provide eventname, subid and function to correctly subscribe")
	end

	cge.__events[eventname][subid] = func
end

function cgevents.unsubscribe(cge, eventname,subid)
	if cge.__events[eventname] == nil then
		error("Event "..eventname.." does not exist")
	end

	if subid == nil or func == nil then
		error("user must provide eventname and subidto correctly unsubscribe")
	end
	
	cge.__events[eventname][subid] = nil
end

function cgevents.publish(cge, eventname,args)
	if cge.__events[eventname] == nil then
		error("Event "..eventname.." does not exist")
	end

	for subid, func in pairs(cge.__events[eventname]) do
		func(unpack(args))
	end
end

rawset(_G,"cgevents", cgevents)

-- base 64 libraries

local b64chars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'

local function b64ify(boolValues)
    local bits = {}
    for i = 1, #boolValues do
        bits[#bits + 1] = boolValues[i] and 1 or 0
    end

    local base64 = ''
    for i = 1, #bits, 6 do
        local byte = 0
        for j = 0, 5 do
            byte = byte * 2
            if bits[i + j] then
                byte = byte + bits[i + j]
            end
        end
        base64 = base64 .. b64chars:sub(byte + 1, byte + 1)
    end

    return base64
end

local function deb64ify(base64String)
    local bits = {}
    for i = 1, #base64String do
        local char = base64String:sub(i, i)
        local index = b64chars:find(char) - 1
		nvdbp("char",char,"idx",index)
        for j = 5, 0, -1 do
            bits[#bits + 1] = index >= 2^(j)
            index = index % 2^(j)
        end
    end

	-- nvdbp(table.concat(bits,","))
    local boolValues = {}
    for i = 1, 1035 do
        boolValues[i] = bits[i] == true
    end
    return boolValues
end

rawset(_G,"b64ify",b64ify)
rawset(_G,"deb64ify",deb64ify)

-- COM_AddCommand("test_b64ify",function()
-- 	local expected = {}
-- 	local expstr = ""
-- 	for i = 1, 1035 do
-- 		expected[i] = (NV_M_RandomRange(1,100) > 50)
-- 		expstr = $..ternary(expected[i],"1","0")
-- 	end

-- 	local encoded = b64ify(expected)
-- 	local actual = deb64ify(encoded)

-- 	local actstr = ""
-- 	for i = 1, 1035 do
-- 		-- print(tostring(actual[i]))
-- 		actstr = $..ternary(actual[i],"1","0")
-- 	end

-- 	print("Expected: "..expstr)
-- 	print("Converted: "..encoded)
-- 	print("Actual: "..actstr)
-- 	-- print("Result: "..tostring(expected == actual))
-- 	print("Result: "..tostring(expstr == actstr))
-- end)