-- cganalytics.lua
-- Take note of who votes for what, and rtv's
-- hopefully can be used to track map preference trends

-- check for existence of dependencies
-- if nvutils_loaded == nil then
-- 	print("\133WARNING: \130required dependency nvutils.lua not loaded")
-- end

local nv_enablePopTrack = CV_RegisterVar({
	name = "nv_enablePopTrack",
	defaultvalue = "On",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

local nv_include_total_state_in_tracking = CV_RegisterVar({
	name = "nv_include_total_state_in_tracking",
	defaultvalue = "On",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

local RECORD_FILE = "kartevents.txt"

-- schema for records
local cgschema = {}
cgschema["0.1"] = {
		schemaver = 1,
		eventtype = 2,
		eventid = 3,
		timestamp = 4,
		servername = 5,
		name = 6,
		skin = 7,
		color = 8,
		speed = 9,
		weight = 10,
		speedweight = 11,
		playerchoiceidraw = 12,
		playerchoicefullraw = 13,
		playerchoiceid = 14,
		playerchoicefull = 15,
		playerchoicewasencore = 16,
		playerchoicenum = 17,
		playerwasafk = 18,
		choicesids = 19,
		choicesfullform = 20,
		choiceencorecapacities = 21
	}

	cgschema["0.2"] = {
		schemaver = 1,
		eventtype = 2,
		eventid = 3,
		timestamp = 4,
		servername = 5,
		name = 6,
		skin = 7,
		color = 8,
		speed = 9,
		weight = 10,
		speedweight = 11,
		playerchoiceidraw = 12,
		playerchoicefullraw = 13,
		playerchoiceid = 14,
		playerchoicefull = 15,
		playerchoicewasencore = 16,
		playerchoicenum = 17,
		playerwasafk = 18,
		choicesids = 19,
		choicesfullform = 20,
		choiceencorecapacities = 21,
		wantsrtv = 22,
		isinitiator = 23
	}

	cgschema["0.3"] = {
		schemaver = 1,
		eventtype = 2,
		eventid = 3,
		timestamp = 4,
		servername = 5,
		name = 6,
		skin = 7,
		color = 8,
		speed = 9,
		weight = 10,
		speedweight = 11,
		playerchoiceidraw = 12,
		playerchoicefullraw = 13,
		playerchoiceid = 14,
		playerchoicefull = 15,
		playerchoicewasencore = 16,
		playerchoicenum = 17,
		playerwasafk = 18,
		choicesids = 19,
		choicesfullform = 20,
		choiceencorecapacities = 21,
		choicepoolwaspreferred = 22,
		wantsrtv = 23,
		isinitiator = 24,
		fulldeliststate = 25,
		fullhellstate = 26,
		fullpreferredstate = 27
	}

local SCHEMA_VERSION = "0.3"

local function initrecording(path,schemaver)
	local fd = io.open(path,'r')
	local fileexists = fd ~= nil
	if fileexists then
		io.close(fd)
	else
		-- start a file with the csv headers
		local headers = {}
		local schema = cgschema[schemaver]
		for k,col in pairs(schema) do
			headers[col] = k
		end
		local headers_str = table.concat(headers,",")
		local fd2 = io.open(path,'a')
		fd2:write(headers_str.."\n")
		io.close(fd2)
	end
end

local function writerecord(path,record,schemaver)
	-- nvdbp("writing record to "..path)
	if not isserver return end

	initrecording(path,schemaver)

	local fd = io.open(path,'a+')
	-- convert table to just its values, comma separated
	local concattable = {}
	local schema = cgschema[schemaver]
	for k,v in pairs(record) do
		local col = schema[k]
		concattable[col] = v

	end
	local output = table.concat(concattable,",").."\n"
	fd:write(output)
	fd:flush()
	fd:close()
	nvdbp("wrote record to "..path)
end

local function NV_recordRtv(p,wantsrtv,isinitiator)
	if not nv_enablePopTrack.value then return end
	local currenttime = cg_get_current_time()

	local target, isencore = nil, nil
	if server.nv_result != nil then
		-- players know which map is coming up next. If they 
		-- rtv now, it's in protest of the upcoming map
		local mapinfo = get_nv_choices()[server.nv_result] 
		target, isencore = mapinfo[1], mapinfo[2]
	else
		-- player just straight up doesn't like current map
		target, isencore = gamemap, encoremode
	end
	isencore = ternary(isencore,1,0)
	local wants_private = nvs_get(p.nvsettings,"nv_donottrackme")
	nvdbp("Player "..tostring(p).." "..tostring(wants_private))
	-- nvdbp("is encore", isencore)
	local record = {
		schemaver = SCHEMA_VERSION,
		eventtype = "rtv",
		eventid = hashString(CV_FindVar("servername").string).."|"..currenttime,
		timestamp = currenttime, 
		servername = CV_FindVar("servername").string,
		name = ternary(wants_private,"[REDACTED]",string.gsub(p.name,",","<comma>")), 
		skin = p.mo.skin,
		color = ternary(wants_private,"[REDACTED]",p.skincolor),
		speed = p.kartspeed,
		weight = p.kartweight,
		speedweight = tostring(p.kartspeed..p.kartweight),
		playerchoiceidraw = target,
		playerchoicefullraw =  build_map_title(mapheaderinfo[target]),
		playerchoiceid = "",
		playerchoicefull = "",
		playerchoicewasencore = isencore,
		playerchoicenum = "",
		playerwasafk =  "",
		choicesids = "",
		choicesfullform = "",
		choiceencorecapacities = "",
		wantsrtv = ternary(wantsrtv,1,0),
		isinitiator = ternary(isinitiator,1,0),
		choicepoolwaspreferred = "",
		fulldeliststate = "",
		fullhellstate = "",
		fullpreferredstate = ""
	}
	writerecord(RECORD_FILE,record,SCHEMA_VERSION)

end
rawset(_G,"NV_recordRtv", NV_recordRtv)

local function recordVotes()
	if not nv_enablePopTrack.value then return end
	local currenttime = cg_get_current_time()
	local ballotid = hashString(CV_FindVar("servername").string).."|"..currenttime

	local i = 0
	for p in players.iterate do
		i=$+1
		if p.spectator then continue end

		local rawchoiceids, choiceencores = {}, {}
		local nvchoices = get_nv_choices()
		for k,v in ipairs(nvchoices) do
			table.insert(rawchoiceids,v[1])
			-- apparently cannot concat booleans
			if v[2] then 
				table.insert(choiceencores,1)
			else
				table.insert(choiceencores,0)
			end
		end
		local wasencore = 0
		if nvchoices[p.nv_choice][2] then wasencore = 1 end
		local wants_private = nvs_get(p.nvsettings,"nv_donottrackme")

		local fulldlstate, fullhstate, fullpstate = "", "", ""
		if nv_include_total_state_in_tracking.value then
			fulldlstate = ternary(server.nv_deliststate == nil,"<error>",server.nv_deliststate)
			fullhstate = ternary(server.nv_hellstate == nil,"<error>",server.nv_hellstate)
			fullpstate = ternary(server.nv_preferredstate == nil,"<error>",server.nv_preferredstate)
		end

		local record = {
			schemaver = SCHEMA_VERSION,
			eventtype = "ballotcast",
			eventid = tostring(ballotid.."::"..i),
			timestamp = currenttime, 
			servername = CV_FindVar("servername").string,
			name = ternary(wants_private,"[REDACTED]",string.gsub(p.name,",","<comma>")),
			skin = p.mo.skin,
			color = ternary(wants_private,"[REDACTED]",p.skincolor),
			speed = p.kartspeed,
			weight = p.kartweight,
			speedweight = tostring(p.kartspeed..p.kartweight),
			playerchoiceidraw = nvchoices[p.nv_choice][1],
			playerchoicefullraw =  build_map_title(mapheaderinfo[nvchoices[p.nv_choice][1]]),
			playerchoiceid = ternary(p.nv_choice == 4,0,nvchoices[p.nv_choice][1]),
			playerchoicefull = ternary(p.nv_choice == 4,"RANDOM",build_map_title(mapheaderinfo[nvchoices[p.nv_choice][1]])),
			playerchoicewasencore = wasencore,
			playerchoicenum = p.nv_choice, -- to detect if they chose random
			playerwasafk =  p.nv_wasafk, -- if random, either afk or disliked choices
			choicesids = table.concat(rawchoiceids,"|"), -- I don't want to make a separate sql table for ballots, let user make em
			choicesfullform = mapsfullform(rawchoiceids),
			choiceencorecapacities = table.concat(choiceencores,"|"),
			wantsrtv = "",
			isinitiator = "",
			choicepoolwaspreferred = ternary(server.nv_poolwaspreferred == nil,"false",tostring(server.nv_poolwaspreferred)),
			fulldeliststate = fulldlstate,
			fullhellstate = fullhstate,
			fullpreferredstate = fullpstate
		}

		-- print(type(record['eventid']))

		writerecord(RECORD_FILE,record,SCHEMA_VERSION)
	end
end
rawset(_G,"NV_recordVotes", recordVotes)
