print("Loading cgtime...")
-- server owners need to insert this line into their launch script:
-- windows: powershell [long]((date).touniversaltime()-[datetime]'1970-01-01').totalseconds > /kartpath/luafiles/currentutc.txt
-- Linux: date +%s > /kartpath/luafiles/currentutc.txt

local nv_enableTime = CV_RegisterVar({
	name = "nv_enableTime",
	defaultvalue = "On",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

-- tracking the current time
local currentUTC = nil
local starttime = nil
local framecounter = 0
local UTC_FILE = "currentutc.txt"
local PINGFILE = "ping.txt"


local function get_current_time()
	return currentUTC
end

local function set_current_time(time)
	if time == nil then return end
	currentUTC = time
end

COM_AddCommand("__set_cgtime",function(p,time) 
	local t = tonumber(time)
	if t == nil then return end
	nvdbp("setting current time to "..t.." (utc)")
	set_current_time(t)
end,COM_ADMIN)

local function readTimeFile(filename)
	if filename == nil then
		filename = UTC_FILE
	end

	local fd = io.open(filename,"r")
	if fd == nil then

		-- stfu if user said they don't want to see this. We always need to init, but nothing after that
		if nv_enableTime.value and CV_FindVar("nv_enableNV").value then
			print("Warning: Could not determine the current time - "..filename.." not found.")
		end
		return
	end
	local utc = fd:read()
	fd:close()
	return utc
end

local function convertSecToHumanReadable(sec)
	if sec == 0 then
		return "0 seconds"
	end
	local days2sec, hours2sec, min2sec = 86400, 3600, 60

	local remaining = sec
	local days = remaining / days2sec

	remaining = remaining - (days * days2sec)

	local hours = remaining / hours2sec

	remaining = remaining - (hours * hours2sec)

	local minutes = remaining / min2sec
	
	remaining = remaining - (minutes * min2sec)

	local seconds = remaining

	local labels = {"days", "hours", "minutes", "seconds"}
	local numbers = {days, hours, minutes, seconds}

	local outstr = ""
	for i, label in ipairs(labels) do
		local number = numbers[i]
		if number == 0 then continue end
		outstr = $.." "..number.." "..label
	end

	return outstr
end

cgevents.register(cgevents,"utctimeset")

local function afterSetUTCTime(status,time,diff)
	if not status return end
	nvdbp("Synced up a difference of"..convertSecToHumanReadable(diff))
	nvdbp("Synced up a difference of "..diff)
	cgevents.publish(cgevents,"utctimeset",{time,diff})
end

local function setUTCTime(is_init)
	if not isserver then return false end
	-- we need to init in all cases, else stfu
	if not is_init and not nv_enableTime.value or not CV_FindVar("nv_enableNV").value then 
		return 
	end

	local success = true
	local new_utc = readTimeFile()

	local new_utc_empty = new_utc == "" or new_utc == nil
	if not new_utc_empty then
		new_utc = tonumber(new_utc)
	end

	if not is_init and (new_utc_empty or (new_utc < get_current_time())) then
		-- currentUTC should always be lagging behind, not ahead
		nvdbp("Time is behind or empty...",new_utc)
		success = false
	end
	
	if new_utc_empty and is_init then
		new_utc = 1542359569 -- easter egg!
	end

	if is_init then
		starttime = new_utc
	end

	-- I wish I could use ternary here but the subtraction is eval'd first. nil could appear
	local diff
	if is_init or new_utc_empty then
		diff = 0
	else
		diff = new_utc - get_current_time()
	end

	if success then
		set_current_time(new_utc)
		if server != nil then 
			COM_BufInsertText(server,"__set_cgtime "..new_utc)
		end
	end
	
	afterSetUTCTime(success,new_utc,diff)
	return success
end
setUTCTime(true)

local exec_schedule = {}
local function incrementutctime()
	-- I'm making the assumption that 35 tics will ALWAYS happen in a second
	-- no way to detect otherwise
	-- https://discord.com/channels/270206720897318912/514241367560683520/1079956760909062185
	-- this will drift to being slow. Owners should restart daily
	-- actually!! :)) Servers going to sleep :) Will not :) fire think frame :)))
	framecounter = $+1
	if framecounter == 35 then
		framecounter = 0
		currentUTC = $+1

		-- execute any commands scheduled at this time, if any
		if exec_schedule[currentUTC] != nil then
			local commands = exec_schedule[currentUTC]
			table.remove(exec_schedule,currentUTC)
			for i,cmd in pairs(commands) do
				cmd.func(unpack(cmd.args))
			end
		end
	end
end
-- exec_schedule[sched_time]
addHook("ThinkFrame",incrementutctime)
addHook("VoteThinker",incrementutctime)
addHook("IntermissionThinker",incrementutctime)
local function netvars(net)
	currentUTC = net($)
end

addHook("NetVars", netvars)

local function doPing(path,action)
    if action == nil then action = "ping" end

    local fd = io.open(path,"w+")
    fd:write(action)
	fd:flush()
	fd:close()
    -- nvdbp("wrote to pingfile") 
end

local function catchTimeResync(p,attempts)
	attempts = tonumber(attempts)
	local maxattempts = 2 * TICRATE
	if attempts > maxattempts then
		CONS_Printf(p,"Not that many attempts. Max="..maxattempts)
		return 
	end
	if attempts < 1 then
		nvdbp("Ran out of time resync attempts")
		return
	end
	if not setUTCTime(false) then
		-- print("Trying again "..attempts)
		COM_BufInsertText(server,"__catch_cgt_resync "..attempts - 1)
	end
end
COM_AddCommand("__catch_cgt_resync",catchTimeResync,COM_ADMIN)

local function startTimeResync(source)
	if not nv_enableTime.value or not CV_FindVar("nv_enableNV").value then return end
	nvdbp("cgtime: Attempting resync from "..source)
	doPing(PINGFILE)
	catchTimeResync(server,TICRATE + 5)
end

local serverisidle = false

local function serverSleepCheck()
	if not nv_enableTime.value or not CV_FindVar("nv_enableNV").value then return end
	local playerspresent = false
	for p in players.iterate do
		if p.valid then
			playerspresent = true
			break
		end
	end

	if serverisidle then
		if playerspresent then
			startTimeResync("wake up")
			serverisidle = false
		end
	else
		if not playerspresent then
			serverisidle = true
		end
	end
end

local prev_gs = nil
local GS_GAME, GS_INT, GS_VOTE = 0x1, 0x2, 0x4
addHook("VoteThinker",function() prev_gs = GS_VOTE end)
addHook("ThinkFrame",function()
	serverSleepCheck()
	prev_gs = GS_GAME
end)

addHook("IntermissionThinker",function()
	if prev_gs != GS_INT then
		-- just went into intermission
		startTimeResync("intermission start")
	end
	prev_gs = GS_INT
end)

local function __get_cgtime_frame()
	return framecounter
end

-- gpt also.. god this saved me so much time
local function convertUtcTime(timestamp,timezoneOffset, minuteOffset)
	-- Constants for calculations
	local secondsPerDay = 86400
	local secondsPerHour = 3600
	local secondsPerMinute = 60

	if minuteOffset == nil then
		minuteOffset = 0
	end

	-- Adjust timestamp for timezone offset
	timestamp = timestamp + (timezoneOffset * secondsPerHour) + (minuteOffset * secondsPerMinute)
	
	-- Calculate number of days since Unix epoch (Jan 1, 1970 at 00:00:00 UTC)
	local daysSinceEpoch = timestamp / secondsPerDay
	
	-- Calculate number of seconds since midnight
	local secondsSinceMidnight = timestamp % secondsPerDay
	
	-- Calculate hour, minute, and second values
	local hour = secondsSinceMidnight / secondsPerHour
	local minute = (secondsSinceMidnight % secondsPerHour) / secondsPerMinute
	local second = secondsSinceMidnight % secondsPerMinute
	
	-- Calculate year, month, and day values
	local year = 1970
	local month = 1
	local day = 1
	while daysSinceEpoch >= 365 do
		if year % 4 == 0 and (year % 100 ~= 0 or year % 400 == 0) then
		daysSinceEpoch = daysSinceEpoch - 366
		year = year + 1
		else
		daysSinceEpoch = daysSinceEpoch - 365
		year = year + 1
		end
	end
	while daysSinceEpoch > 0 do
		local daysInMonth = 31
		if month == 2 then
		if year % 4 == 0 and (year % 100 ~= 0 or year % 400 == 0) then
			daysInMonth = 29
		else
			daysInMonth = 28
		end
		elseif month == 4 or month == 6 or month == 9 or month == 11 then
		daysInMonth = 30
		end
		if daysSinceEpoch >= daysInMonth then
		daysSinceEpoch = daysSinceEpoch - daysInMonth
		month = month + 1
		if month > 12 then
			month = 1
			year = year + 1
		end
		else
		day = day + daysSinceEpoch
		daysSinceEpoch = 0
		end
	end
	
	return {
		year=year,
		month=month,
		day=day,
		hour=hour,
		minute=minute,
		second=second
	}
  end

local function valid_timestr(offsetstr)
	local components = pipesplit(offsetstr,":")
	local hour, minute = components[1], components[2]

	if hour == nil then
		return {
			receipt = nil,
			reason = "Hour component not readable"
		}		
	end
	if minute == nil then
		minute = 0
	end

	hour = tonumber(hour)
	minute = tonumber(minute)

	if (hour == nil) or (minute == nil) then
		return {
			receipt = nil,
			reason = "Minute or Hour component is not a number"
		}
	end

	if hour < -12 or hour > 12 then
        return {
            receipt = nil,
            reason = "Hour component must be within range [-12, 12]."
        }
    end

	if minute < 0 or minute > 59 then
        return {
            receipt = nil,
            reason = "Minute component must be within range [0, 59]."
        }
    end

	return {
		receipt = offsetstr,
		reason = "success"
	}
end
rawset(_G,"valid_timestr",valid_timestr)

local function interpret_offset(offsetstr)
	local valid_result = valid_timestr(offsetstr)
	if valid_result.receipt == nil then
		error("offset string "..offsetstr.." not valid. Reason: "..valid_result.reason)
	end

	local components = pipesplit(offsetstr,":")
	local hour, minute = components[1], components[2]
	if minute == nil then
		minute = 0
	end

	hour = tonumber(hour)
	minute = tonumber(minute)
	local firstchar = lstrip(offsetstr):sub(1,1)
	-- nvdbp(offsetstr,lstrip(offsetstr),firstchar == "-")
	if hour < 0 or firstchar == "-" then
		minute = $ * -1
	end
	return {hour=hour,minute=minute}
end

local function cgt_fulltime_readout(timestamp,timezoneOffset)
	local interpreted = interpret_offset(timezoneOffset)
	local time = convertUtcTime(timestamp,interpreted.hour,interpreted.minute)
	-- Format date and time string
	local formattedTime = string.format("%04d%02d%02d::%02d%02d%02d", time.year, time.month, time.day, time.hour, time.minute, time.second)

	return formattedTime
end

local function cgt_condensed_readout(timestamp, timezoneOffset, ampm)
	local interpreted = interpret_offset(timezoneOffset)
	local time = convertUtcTime(timestamp,interpreted.hour,interpreted.minute)

	local ampmtxt = ""
	if ampm then
		ampmtxt = ternary(time.hour < 12, "AM", "PM")
		if time.hour == 0 then
			time.hour = 12
		elseif time.hour > 12 then
			time.hour = $-12
		end
	end
	-- Format date and time string
	local formattedTime = string.format("%d:%02d%s", time.hour, time.minute,ampmtxt)
	return formattedTime

end

local function schedule_seconds(func,args,seconds)
	local sched_time = currentUTC + seconds
	local commands = exec_schedule[sched_time]
	if commands == nil then commands = {} end
	table.insert(commands,{func=func,args=args})
	exec_schedule[sched_time] = commands
	print("scheudling time at "..sched_time)
end
rawset(_G,"cg_get_current_time", get_current_time)
rawset(_G,"cgt_fulltime_readout", cgt_fulltime_readout)
rawset(_G,"cgt_condensed_readout", cgt_condensed_readout)
rawset(_G,"__get_cgtime_frame", __get_cgtime_frame)
rawset(_G,"cg_schedule_func_sec",schedule_seconds)
