print("Loading NuVote...")
rawset(_G,"nuvote",{})
nuvote.VERSION = 0
nuvote.PATCH = 53

local nv_enableNV = CV_FindVar("nv_enableNV")

local nv_honor_preempted_votes = CV_RegisterVar({
	name = "nv_honor_preempted_votes",
	defaultvalue = "On",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

local showSMR = CV_RegisterVar({
	name = "nv_showsmr",
	defaultvalue = "On",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

-- what % of people need to vote for a map for it to be selected?
local nv_supermajoritypct = CV_RegisterVar({
	name = "nv_supermajoritypct",
	defaultvalue = 75,
	flags = CV_NETVAR,
	possiblevalue = {MIN = 51, MAX = 100}
})

local nv_superminoritypct = CV_RegisterVar({
	name = "nv_superminoritypct",
	defaultvalue = 20,
	flags = CV_NETVAR,
	possiblevalue = {MIN = 0, MAX = 50}
})


-- Should all people voting random induce a hell map?
local nv_random_induces_hell = CV_RegisterVar({
	name = "nv_random_induces_hell",
	defaultvalue = "On",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

--  How often should hell maps show up as normal choices?
local nv_hell_as_nonrandom_choice_pct = CV_RegisterVar({
	name = "nv_hell_as_nonrandom_choice_pct",
	defaultvalue = 0,
	flags = CV_NETVAR,
	possiblevalue = {MIN = 0, MAX = 100}
})

-- How often should hell maps show up as the random choice?
local nv_hell_as_random_choice_pct = CV_RegisterVar({
	name = "nv_hell_as_random_choice_pct",
	defaultvalue = 1,
	flags = CV_NETVAR,
	possiblevalue = {MIN = 0, MAX = 100}
})

-- How often should hell maps show up as the random choice?
local nv_hell_as_randommap_pct = CV_RegisterVar({
	name = "nv_hell_as_randommap_pct",
	defaultvalue = 1,
	flags = CV_NETVAR,
	possiblevalue = {MIN = 0, MAX = 100}
})

-- How often should encore rounds (rounds where 3 are encore) show up? 
local nv_encore_round_pct = CV_RegisterVar({
	name = "nv_encore_round_pct",
	defaultvalue = 0,
	flags = CV_NETVAR,
	possiblevalue = {MIN = 0, MAX = 100}
})

-- How often should single encore maps show up?
local nv_encore_choice_pct = CV_RegisterVar({
	name = "nv_encore_choice_pct",
	defaultvalue = 5,
	flags = CV_NETVAR,
	possiblevalue = {MIN = 0, MAX = 100}
})

local nv_show_vote_bg = CV_RegisterVar({
	name = "nv_show_vote_bg",
	defaultvalue = "Off",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

local nv_play_vote_music = CV_RegisterVar({
	name = "nv_play_vote_music",
	defaultvalue = "Off",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

-- A preferred map should show up every how many voting rounds?
local nv_preferred_every = CV_RegisterVar({
	name = "nv_preferred_every",
	defaultvalue = 20,
	flags = CV_NETVAR | CV_CALL,
	possiblevalue = CV_Unsigned,
	func = function(cv)
		if nuvote["prev_preferred_every"] and nuvote["prev_preferred_every"] > cv.value then
			nvdbp("Resetting match num because admin wants preferred more frequently")
			nuvote["match_num"] = 0
		end
	end
})

-- When a voting round includes preferred maps, how many should there be?
local nv_preferred_ct = CV_RegisterVar({
	name = "nv_preferred_ct",
	defaultvalue = 2,
	flags = CV_NETVAR,
	possiblevalue = {MIN=0, MAX=4}
})

-- the choices
-- local maps, battle, hell, weights = {}, {}, {}, {}
local maps, battle, hell, bhell, pmaps, pbmaps = {}, {}, {}, {}, {}, {}
local delisted_maps = Set.new()
local delist_reasons = {}
local preferred_maps = Set.new()
local hellified_maps = Set.new() -- maps which were moved to hell
local unhellified_maps = Set.new() -- maps which were moved out of hell
local map_cache = {} -- In memory version of NuVoteConfig.txt
-- btw, I'm writing this line very late in development. I can see I can just use this 
-- for all config state instead of three separated bloated 'Sets'. Not gonna bother
-- to 'fix' though. It's November 2023, and it's unlikely to matter


-- initialize maps for selection. We actually have to do this for each
-- vote, since a soc could be loaded at any time
local function mapIsHell(mapid)
	-- old logic
	-- local map = mapheaderinfo[mapid]
	-- return map.menuflags & LF2_HIDEINMENU
	-- Hell logic - if NuVote has any opinion on it, use it, else fallback to headerinfo
	local mapid_num = tonumber(mapid)
	mapid = ternary(mapid_num == nil,mapid,mapid_num)
	local ishell = false
	if Set.contains(hellified_maps,mapid) or Set.contains(unhellified_maps,mapid) then
		ishell = Set.contains(hellified_maps,mapid) and not Set.contains(unhellified_maps,mapid)
	else
		ishell = (mapheaderinfo[mapid].menuflags & LF2_HIDEINMENU) != 0
	end
	return ishell
end

local function loadMaps()
	-- reset maps
	maps, battle, hell, bhell, pmaps,pbmaps = {}, {}, {}, {}, {}, {}
	for i = 0, #mapheaderinfo do
		local map = mapheaderinfo[i]
		-- don't put in any maps that are not loaded
		if not map then continue end
		if Set.contains(delisted_maps,i) then continue end
		if Set.contains(hellified_maps,i) and Set.contains(unhellified_maps,i) then
			print("Warning: "..G_BuildMapName(i).." is considered both a hell and non-hell map")
		end

		local ishell = mapIsHell(i)
		local ispreferred = Set.contains(preferred_maps,i)
		-- A map could be hell, normal, or preferred
		-- A map cannot be both hell and preferred. 1) It's a bit oxymoronic 2) Use case is very small 3) Will confuse owners imo
		-- I could be convinced otherwise
		if map then
			if (map.typeoflevel & TOL_RACE) then
				if ishell then
					table.insert(hell, i)
				else
					table.insert(maps, i)
				end
				if ispreferred then table.insert(pmaps, i) end

			elseif (map.typeoflevel & TOL_MATCH)
				if ishell then
					table.insert(bhell, i)
				else
					table.insert(battle, i)
				end
				if ispreferred then table.insert(pbmaps, i) end
			end
		end
	end

end
loadMaps()

local DEFAULT_EXCLUDE_LAST = 40

local nv_recent_maps = QueuedKeys.new(DEFAULT_EXCLUDE_LAST) -- size defined and set by CVARS

local function check_exclude_last_cv(desired_size)
	local max_maps = #maps + #hell - delisted_maps.size - 5
	print(#maps.." "..#hell.." "..delisted_maps.size.." "..max_maps)
	if desired_size.value > max_maps then
		print("Max possible maps to exclude for vote eligibility is "..max_maps)
		CV_Set(desired_size,max_maps)
		return
	end

	QueuedKeys.updatecapacity(nv_recent_maps,desired_size.value)
end

-- How many maps to remember as ineligible because it was played too recent?
local nv_exclude_last_n = CV_RegisterVar({
	name = "nv_exclude_last_n",
	defaultvalue = DEFAULT_EXCLUDE_LAST,
	flags = CV_NETVAR | CV_CALL,
	possiblevalue = CV_Unsigned,
 	func = check_exclude_last_cv -- maps might be 'delisted'
})

function nuvote.checkIfNewSession(time,diff)
	nvdbp("Got utctimeset event "..time.." "..diff)
	if diff > (30 * 60) then -- 30 minutes time lapse
		QueuedKeys.clear(nv_recent_maps)
	end
end

local funcQueue = {}
funcQueue.command_buffer = {}
funcQueue.cmd_cooldown_timer = 0
funcQueue.cmd_cooldown = 0
addHook("ThinkFrame",function()
	funcQueue.cmd_cooldown = max(0,funcQueue.cmd_cooldown_timer-1)
	if #funcQueue.command_buffer == 0 then return end
	if funcQueue.cmd_cooldown > 0 then return end
	local cmdinfo = table.remove(funcQueue.command_buffer,1)
	if cmdinfo[3] == "callback" then
		local func, args = cmdinfo[1], cmdinfo[2]
		if args != nil then
			func(unpack(args))
		else
			func()
		end
	elseif cmdinfo[3] == "stall" then
		local what = 1
	else
		COM_BufInsertText(cmdinfo[1],cmdinfo[2])
	end
	funcQueue.cmd_cooldown_timer = funcQueue.cmd_cooldown
end)

function funcQueue.add_command_to_buffer(p,command)
	table.insert(funcQueue.command_buffer,{p,command})
end

function funcQueue.add_command_batch(commands,callback)
	for i,cmd in ipairs(commands) do
		table.insert(funcQueue.command_buffer,cmd)
	end
	local func, args = callback[1],callback[2]
	table.insert(funcQueue.command_buffer,{func,args,"callback"})
end

function funcQueue.add_callback_batch(callbacks,final_callback)
	for i, cbk in ipairs(callbacks) do
		local func, args = cbk[1], cbk[2]
		table.insert(funcQueue.command_buffer,{func,args,"callback"})
	end
	if final_callback != nil then
		local func, args = final_callback[1], final_callback[2]
		table.insert(funcQueue.command_buffer,{func,args,"callback"})
	end
end

function funcQueue.add_stall_frames(frames)
	for i=1,frames do
		table.insert(funcQueue.command_buffer,{nil,nil,"stall"})
	end
end

local scalar = 30

local math

local curx = 0
local cury = 0
nuvote.cur1 = 0
nuvote.cur2 = 0
nuvote.cur3 = 0
nuvote.cur4 = 0
nuvote.cur5 = 0
nuvote.cur6 = 0
nuvote.cur7 = 0
nuvote.cur8 = 0
nuvote.cur9 = 0

-- NuVote State
rawset(_G,"NV_DEFAULT",0x0)
rawset(_G,"NV_CHOICE", 0x1)
rawset(_G,"NV_ROULETTE",0x2)
rawset(_G,"NV_RESULT", 0x4)
rawset(_G,"NV_WAIT",0x8)
rawset(_G,"NV_NEXT",0x16)

local nv_state = NV_DEFAULT
local prevState = -1
local nv_animation = nil

rawset(_G,"get_nv_state",function()
	return nv_state
end)

-- Gamestate
nuvote.GS_GAME = 0x1
nuvote.GS_INT = 0x2
nuvote.GS_VOTE = 0x4
nuvote.prev_gamestate = 0
nuvote.intermission_countdown = 0
nuvote.nuvote_preempted = false

-- more state
nuvote.mapwarpcasted = false

-- reactions to a map being picked
local mapReactions = {}
mapReactions[972]="It's a jungle, brother"

-- debug
local huddebug = false


local nv_choices = {value="||||"}
cgs_registerState(nv_choices,"nv_choices")

-- load map reactions
local function loadMapReactions()
	return
end

local function query_map(query_arr)
	-- much of this yoinked from freeman and Dr_Nope maplist listmap utility
	-- build a list of every map
	-- filter out maps that DON"T have each word in provided query_arr
	local max= (36*26+26)+100
	local maptable = {}
	-- we loop thru again in case maps were loaded after this lua is
    for i=0,max do
        local map= mapheaderinfo[i]
        if map and map.lvlttl then
			local is_recent = QueuedKeys.timetoejected(nv_recent_maps,map)
			if is_recent == -1 then is_recent = "NR" end
			table.insert(maptable,
						{id=i,
						extid = G_BuildMapName(i),
						name=build_map_title(map),
						sprint=map.levelflags&LF_SECTIONRACE,
						gametype=map.typeoflevel,
						-- hell=map.menuflags & LF2_HIDEINMENU,
						hell=mapIsHell(i),
						isdelisted=Set.contains(delisted_maps,i),
						delistreason=delist_reasons[i],
						isrecent=is_recent,
						ispreferred=Set.contains(preferred_maps,i)	
						}
					)
        end
    end

	local direct_query = ""
	if #query_arr == 1 then
		if #query_arr[1] == 2 then
			direct_query = "map"..query_arr[1]
		end

		if #query_arr[1] == 5 and string.lower(string.sub(query_arr[1],1,3)) == "map" then
			direct_query = query_arr[1]
		end
	end

	if #direct_query > 0 then
		-- querying by mapid. I am allowing a lotta inefficiency.. but max ct is 1035 so whatevs
		for i=#maptable,1,-1
			if not string.find(string.lower(maptable[i].extid),string.lower(direct_query)) then table.remove(maptable, i) end
		end
	else
		-- querying with a string
		for i=1,#query_arr
			local testarg = string.lower(query_arr[i])
			for i=#maptable,1,-1
				if not string.find(string.lower(maptable[i].name),testarg) then table.remove(maptable, i) end
			end
		end
	end

	return maptable
end

nuvote.mapwasqueued = 0
local function maps_are_queued()
	local commandq_queued = (cq_hasmaps != nil and cq_hasmaps() and CV_FindVar("cq_poponvote").value == 1)
	local ato_queued = (aTourney != nil and #aTourney.tourneyMapList)
	local result = commandq_queued or ato_queued
	if result then
		nuvote.mapwasqueued = 5
	end
	return result
end

local function castVote(player,choice)
	if nv_state != NV_CHOICE then
		chatprintf(player,"\133[PM]<~Votekeeper> Now is not the time to vote.")
		return
	end

	if player.spectator then
		chatprintf(player,"\133[PM]<~Votekeeper> You must first join the game in order to vote. (You are currently spectating)")
		return
	end

	if player.nv_votecooldown > 0 then
		chatprintf(player,"\133[PM]<~Votekeeper> Please wait "..player.nv_votecooldown / TICRATE.." seconds before recasting your vote.")
		return
	end

	NV_playsound(nil,sfx_kc4a,player)
	player.nv_wasafk = 0
	player.nv_choice = choice
	player.nv_votecooldown = 3 * TICRATE
end


-- Players can vote thru chat
local function handlePlayerVote(player,msgType,target,msg)
	if msgType != 0 then return false end
	local validVotes = {'1','2','3','4'}
	if nv_state != NV_CHOICE or msgType != 0 then return end
	
	for i,vv in pairs(validVotes) do
		if msg == vv then
			castVote(player,tonumber(msg)) 
			return true
		end
	end
	return false
end
addHook("PlayerMsg",handlePlayerVote)

-- reset player choice if they spectate
local function checkForSpec()
	for p in players.iterate do
		if p.nv_prevspec and p.nv_prevsepc != p.spectator and p.spectator then
			-- player just turned into a spectator
			-- reset their choice and vote cooldown
			p.nv_choice = 0
			p.nv_votecooldown = 0
		end
		p.nv_prevspec = p.spectator
	end
end

-- the ALgorIThiM!!!!

local function binarySearch(arr,left,right,target)
	while left != right do
		local mid = left + FixedFloor((left - right) / 2)
		if target > arr[mid] then left = mid + 1
		elseif target < arr[mid] then right = mid
		else return arr[mid] end
	end
	
	return arr[left]
end

nuvote.nomap_fallback = 0
local function chooseFrom(choices,backup,poolname,bkp_poolname)
	local pool = choices
	if #choices == 0 then
		if #backup == 0 then
			local msg = string.format("\133Aieee!! Both %s and %s map lists were empty!!. \128Check your NuVoteConfig.txt and soc files, as all maps of these kinds are delisted, socc'd or otherwise unavailable. Giving MAP01,MAP02...",poolname, bkp_poolname)
			print(msg)
			local fallback_map = nuvote.nomap_fallback + 1
			nuvote.nomap_fallback = ($+1) % 5
			return fallback_map
		end
		pool = backup
		local msg = "We tried pulling a "..poolname.." map for voting, but there were no maps configured as such. Check your NuVoteConfig.txt, or soc files. A "..bkp_poolname.." map will be used instead..."
		nvwarn(server,msg)
	end
	local choice = pool[NV_RandomRange(1, max(1,#pool))]
	return choice
end

local function mapNotEligible(mapid,isbattle)
	-- maps are not eligible if we recently played them
	-- Recent could mean during a session, or at least ~ 5 races ago
	-- need to maintain state of recently played maps, as well as
	-- the last completed race (session boundaries)

	local TOL_mask = ternary(isbattle,TOL_MATCH,TOL_RACE)

	local map_soccd = not (mapheaderinfo[mapid].typeoflevel & TOL_mask)
	local map_delisted = Set.contains(delisted_maps,mapid)
	local map_played_recently = QueuedKeys.contains(nv_recent_maps,mapid)
	return map_delisted or map_played_recently or map_soccd
end

nuvote.match_num = -1
local function getChoicePoolType(choicetype,isbattle,choiceid)
	-- Determine which choice pool to pull from: Normal, hell, or preferred.
	-- Ask longon for sprite to indicate hell
	local choiceIsHell = false
	if choicetype == "open" then
		choiceIsHell = NV_RandomChance(FRACUNIT * nv_hell_as_nonrandom_choice_pct.value / 100)
	elseif choicetype == "random" then
		choiceIsHell = NV_RandomChance(FRACUNIT * nv_hell_as_random_choice_pct.value / 100)
	elseif choicetype == "randommap" then
		choiceIsHell = NV_RandomChance(FRACUNIT * nv_hell_as_randommap_pct.value / 100)
	end

	-- determine whether to use preferred pool
	-- Conditions:
	-- * Preferred is enabled (prerred every, preferred count)
	-- * It's actually a round where we need to serve a preferred map
	-- * There are enough preferred maps eligible to serve out
	local maps_on_ballot = choiceid - 1
	local eligible_preferred = 0
	local preferred_pool = ternary(isbattle,pbmaps,pmaps)
	for i,mapid in ipairs(preferred_pool) do
		if mapNotEligible(mapid,isbattle) then continue end
		eligible_preferred = $ + 1
	end
	eligible_preferred = $ - maps_on_ballot
	local preferred_per_round = (CV_FindVar("nv_preferred_ct").value)
	local preferred_remaining = min(preferred_per_round - maps_on_ballot, eligible_preferred)
	local preferred_every = (CV_FindVar("nv_preferred_every").value)
	local usePreferred = (
		(preferred_every > 0) and 
		(preferred_remaining > 0) and 
		((nuvote.match_num % preferred_every) == 0)
	)
	if usePreferred then
		server.nv_poolwaspreferred = true
		return "preferred"
	elseif choiceIsHell then
		return "hell"
	else
		return "normal"
	end
end


local function checkExcludeLastList()
	-- just set nv_exclude_last_n again to trigger check for excluded maps
	-- so we are sure to not hit a gbj caused by midruntime map exclusions

	CV_Set(nv_exclude_last_n,nv_exclude_last_n.value)
end

local function get_nv_choices()
	-- local raw = nv_choices.string
	local raw = cgs_get("nv_choices")
	-- print(raw)
	local choices_str = pipesplit(raw,"|")
	if #choices_str == 0 then return nil end
	local choices = {}
	for i,choice in ipairs(choices_str) do
		local choice_arr = pipesplit(choice,",")
		local map, is_encore = choice_arr[1], choice_arr[2]
		is_encore = ternary(is_encore == "1",true,false)
		choices[i] = {map,is_encore}
	end
	return choices
end
rawset(_G,"get_nv_choices", get_nv_choices)


local function set_nv_choices(choices)
	local choices_str = {}
	for i,choice in ipairs(choices) do
		local map, is_encore = choice[1], choice[2]
		is_encore = ternary(is_encore,"1","0")
		choices_str[i] = map..","..is_encore
	end

	local value = table.concat(choices_str,"|")

	if not isserver then return end
	-- CV_Set(nv_choices,value)
	cgs_sendState("nv_choices",value)
	-- nv_choices.value = value
	cgs_registers["nv_choices"].value = value
end

local function updatePrngWithTimes()
	local ptimes = ""
	for p in players.iterate do
		if p.spectator then
			ptimes = $.."S"
		else
			ptimes = $..tostring(p.realtime)
		end
	end
	-- nvdbp("setting ptime",ptimes)
	update_prng_txt(ptimes)
end

function nuvote.tunes(musictag)
	for p in players.iterate do
		COM_BufInsertText(p, "tunes "..musictag)
	end
end

local kartpositions = {}
local function saveKartPositions()
	kartpositions = {}
	for i=0,(#players-1) do
		local pos
		if players[i] == nil then
			pos = nil
		else
			pos = players[i].kartstuff[k_position]
		end
		kartpositions[i] = pos
	end
end

local function restoreKartPosition(p)
	-- get index on table
	local pidx
	for i=0,(#players-1) do
		if players[i] == p then
			pidx = i
			break
		end
	end
	local prev_pos = kartpositions[i]
	if (prev_pos == nil) or (prev_pos == 0) then return end
	players[i].kartstuff[k_position] = prev_pos -- i hope this owrks
end


function nuvote.randommap_helper(choicetype,choiceid)
	if choiceid == nil then
		choiceid = 1
	end
	
	-- Chooses one random map according to current state
	local isbattle = gametype == 1
	local poolType = getChoicePoolType(choicetype,isbattle,choiceid)
	local bkp_pool = ternary(isbattle,battle,maps)
	local bkp_poolname = ternary(isbattle,"Normal battle","Normal")
	local pool_map = {
		["preferred"] = {[true] = pbmaps, [false] = pmaps},
		["hell"] = {[true] = bhell, [false] = hell},
		["normal"] = {[true] = battle, [false] = maps},
	}

	local poolname_map = {
		["preferred"] = {[true] = "Preferred battle", [false] = "Preferred"},
		["hell"] = {[true] = "Battle Hell", [false] = "Hell"},
		["normal"] = {[true] = "Battle", [false] = "Normal"},
	}

	local pool = pool_map[poolType][isbattle]
	local poolname = poolname_map[poolType][isbattle]

	local candidate = chooseFrom(pool,bkp_pool,poolname,bkp_poolname)
	local failsafe = (#pool == 0 and #bkp_pool == 0)
	while mapNotEligible(candidate,isbattle) and not failsafe do
		candidate = chooseFrom(pool,bkp_pool,poolname,bkp_poolname)
	end
	local isencore = NV_RandomChance(FRACUNIT * nv_encore_choice_pct.value / 100)

	return {candidate, isencore}
end

local function loadNewChoices()

	local choices, encores = {}, {}


	-- First choose the maps
	local chosens = {sentinel=true}
	for i=1,4 do
		local candidate, isencore = "sentinel", false
		local choicetype = ternary((i == 4),"random","open")
		while chosens[candidate] ~= nil do
			local candinfo = nuvote.randommap_helper(choicetype, i)
			candidate, isencore = candinfo[1], candinfo[2]
		end
		chosens[candidate] = true
		choices[i] = candidate
		encores[i] = isencore
	end

	local isbattle = gametype == 1
	local hpool = ternary(isbattle,bhell,hell)
	local hpoolname = ternary(isbattle,"Battle hell","Hell")
	local bkp_pool = ternary(isbattle,battle,maps)
	local bkp_poolname = ternary(isbattle,"Battle","Normal")
	local hellChoice = chooseFrom(hpool,bkp_pool,hpoolname,bkp_poolname)


	
	local hellEncore = false
	local encoreEnabled = true
	-- encore Round: The event where 3 of 4 choices are encore
	local encoreRound = NV_RandomChance(FRACUNIT * nv_encore_round_pct.value / 100)
	if encoreRound then
		local notEncoreChoice = NV_RandomRange(1, 4)
		for i=1,4 do
			if i == notEncoreChoice then continue end
			encores[i] = true
		end
	end

	hellEncore = encores[4] -- if hell is chosen, encore capacity should match
	

	-- now populate the netvars
	local nvchoices = {}
	for i=1,4 do
		nvchoices[i] = {choices[i],encores[i]}
	end
	nvchoices[5] = {hellChoice,hellEncore}

	-- print("about to set hoices"..nvchoices)
	set_nv_choices(nvchoices)

	local dlstatearr, hellstatearr, prefstatearr = {}, {}, {}
	for i = 1,1035 do
		local map = mapheaderinfo[i]
		if not map then
			dlstatearr[i] = false
			hellstatearr[i] = false
			prefstatearr[i] = false
			continue
		end
		local map_soccd = not (mapheaderinfo[i].typeoflevel & (TOL_MATCH|TOL_RACE))
		local nv_delisted = Set.contains(delisted_maps,mapid)
		local map_delisted = map_soccd or nv_delisted
		dlstatearr[i] = not map_delisted
		hellstatearr[i] = mapIsHell(i)
		prefstatearr[i] = Set.contains(preferred_maps,i)
	end

	server.nv_deliststate = b64ify(dlstatearr)
	server.nv_hellstate = b64ify(hellstatearr)
	server.nv_preferredstate = b64ify(prefstatearr)
end

function nuvote.test_randommap_helper()
	-- test if randommap_helper works as intended. It's core to the algorithm
	loadMaps()
	CV_Set(nv_hell_as_random_choice_pct,20)
	CV_Set(nv_encore_choice_pct,5)
	CV_Set(nv_exclude_last_n,20)
	for i=1,20 do
		QueuedKeys.add(nv_recent_maps,i)
	end
	-- Normal maps are 1 - 60
	-- Hell maps are maps 352 - 369, inclusive 
	-- maps 1 -20 in this test are recent, so none should appear
	-- Maps 21 - 24 are delisted, inclusive, outside of here
	-- Map 25 is made hell by not soc
	-- Map 26 is preferred
	-- Map 27 is made hell by soc
	-- Gangbare Dochu (map 41) is socced outside of here

	local freqs = {}
	local num_encore = 0
	local total_runs = 50000
	local pct_to_bps = total_runs / 100

	-- calculate expected percentages
	local normal_maps = 60
	local natural_hell_maps = 365 - 352 + 1 + 1
	local total_maps = normal_maps + natural_hell_maps
	local eligible_maps = total_maps - (4 + 1 + 20) -- delisted + soc + recent


	local forced_hell_maps = 2
	local total_hell_maps = natural_hell_maps + forced_hell_maps
	
	-- put stuff here for preferred
	local total_preferred_maps = 0

	local total_normal_maps = eligible_maps - (total_hell_maps + total_preferred_maps)

	-- these are our expected values, but it's a fuzzy expected
	local chosen_hell_maps = (nv_hell_as_random_choice_pct.value * total_runs) / 100
	local chosen_normal_maps = total_runs - chosen_hell_maps
	local exp_encores = (nv_encore_choice_pct.value * total_runs) / 100

	print("Chosen normal maps: "..chosen_normal_maps)
	print("Chosen hell maps: "..chosen_hell_maps)


	local bps_hell_maps = chosen_hell_maps / total_hell_maps
	local bps_normal_maps = chosen_normal_maps / total_normal_maps
	local bps_preferred_maps = bps_normal_maps

	-- print("Total runs")
	print("Total normal maps: "..total_normal_maps)
	print("Total hell maps: "..total_hell_maps)

	for i=1,total_runs do
		local res = nuvote.randommap_helper("open")
		local mapid,encore = res[1], res[2]
		if freqs[mapid] == nil then
			freqs[mapid] = 1
		else
			freqs[mapid] = $+1
		end
		num_encore = $ + ternary(encore,1,0)
	end

	local total_act_f = 0
	local total_exp_f = 0

	for mapid = 1,400 do
		if freqs[mapid] == nil then continue end
		local f = freqs[mapid]
		if mapid == 41 or mapid < 25 or mapid == 1035 then
			error("Map "..mapid.." was selected "..f.." times when it should have never been chosen")
		end

		local expected
		if mapid == 26 then
			expected = bps_preferred_maps
		elseif mapid > 60 or mapid == 25 or mapid == 27 then
			expected = bps_hell_maps
		else
			expected = bps_normal_maps
		end
		total_act_f = $+f
		total_exp_f = $+expected
		print(mapid..": Actual: "..f.." Exp: "..expected.." Diff: "..(f-expected))
	end

	print("T Act: "..total_act_f.." T Exp: "..total_exp_f)
	print("Encores: "..num_encore.." Expected: "..exp_encores)

	return true
end

-- Allow player to vote using controller
local NV_UIChoiceElement = {}
function NV_UIChoiceElement.new(id)
  return {id=id,north=nil,south=nil,east=nil,west=nil}
end
local choice1 = NV_UIChoiceElement.new("choice1")
local choice2 = NV_UIChoiceElement.new("choice2")
local choice3 = NV_UIChoiceElement.new("choice3")
local choice4 = NV_UIChoiceElement.new("choice4")
choice1.south,choice1.east = choice3,choice2
choice2.south,choice2.west = choice4,choice1
choice3.north,choice3.east = choice1,choice4
choice4.north,choice4.west = choice2,choice3

local ui_cursor = choice1

local consoleplayerbus = {}

local function movecursorgenerator(direction)
	local function cursormover(player)
		if nv_state != NV_CHOICE return end
		if player.spectator then return end
		if player.nv_choice != 0 then return end
		-- nvdbp("i'm being called "..direction)
		table.insert(consoleplayerbus,{key="nv_wasafk",value=0})
		if ui_cursor[direction] != nil then
			S_StartSound(nil,sfx_kc4a,player)
			ui_cursor = ui_cursor[direction]
		end
	end
	return cursormover
end

local localprevinput = {
	up = false,
	down = false,
	left = false,
	right = false,
	accel = true,
	accelprev = {},
	acceltime = {},
	ltime = 0,
	accelframes = 0,
	votestart = 0
}

function nuvote.timeid()
	return string.sub(tostring(leveltime),-2)
end

local function confirmselection(player)
	if nv_state != NV_CHOICE then return end
	if player.spectator then return end
	if player.nv_showhelp then return end
	if player.nv_choice != 0 then return end
	if #consoleplayerbus > 0 then return end
	if consoleplayer == nil then return end
	if player != consoleplayer then return end
	nvdbp("Got confrim on frame "..nuvote.timeid())
	local selectnum = string.sub(ui_cursor.id, -1)
	local selection = tonumber(selectnum)
	table.insert(consoleplayerbus,{key="nv_choice",value=selection})
end

function nuvote.reset_localprevinput()
	localprevinput = {
		up = false,
		down = false,
		left = false,
		right = false,
		accel = true,
		accelprev = {},
		acceltime = {},
		ltime = 0,
		accelframes = 0,
		votestart = 0
}
end

local risingEdgeFunctions = {
	up = movecursorgenerator("north"),
	down = movecursorgenerator("south"),
	left = movecursorgenerator("west"),
	right = movecursorgenerator("east"),
	accel = confirmselection,
}


rawset(_G,"BT_UP",32)
rawset(_G,"BT_DOWN",64)

local function handlePlayerInput(player,cmd,triggerSrc)
	if localprevinput == nil then return end
	if player.spectator then return end
	if player.nv_showhelp then return end
	-- nvdbp("pid","fwd","side","buttons")
	-- nvdbp(player.name,cmd.forwardmove,cmd.sidemove,cmd.buttons)
	local encorefactor = ternary(encoremode,-1,1)
	local up = (cmd.buttons & BT_UP) != 0
	local down = (cmd.buttons & BT_DOWN) != 0
	local left = (cmd.driftturn * encorefactor) > 0
	local right = (cmd.driftturn * encorefactor) < 0
	local accel = (cmd.buttons & BT_ACCELERATE) > 0 -- do not accept accel+brake

	local nvcmds = {
		up=up,
		down=down,
		left=left,
		right=right,
		accel=accel
	}


	if localprevinput["ltime"] >= leveltime then
		print("Multiple calls to searchInput on the same frame! Notify cglitcher")
	end

	localprevinput["ltime"] = leveltime

	local accelstr = localprevinput["accelprev"]
	local accelt = localprevinput["acceltime"]
	table.insert(accelstr,accel)
	table.insert(accelt,nuvote.timeid())

	local prev = accelstr[1]
	local gap = ternary(prev,0,1)
	local mingap = gap
	local gaps = {}
	for i = 2,#accelstr do
		local cur = accelstr[i]
		if not prev and cur then
			-- rising edge detected
			mingap = min(mingap,gap)
			gap = 0
			table.insert(gaps,gap)
			if #accelstr == 2 then
				print("Accel was pegged as on for some reason. Notify cglitcher")
			end
		elseif not cur then
			gap = $+1
		end
		prev = cur
	end

	local shortgaps = 0
	for i,gap in ipairs(gaps) do
		if gap < 4 then
			shortgaps = $+1
		end
	end

	if shortgaps > 1 then
		print("Manchester-like signal detected on accel button. Notify cglitcher")
	end

	while #accelstr > 10 do
		table.remove(accelstr,1)
		table.remove(accelt,1)
	end
	localprevinput["accelprev"] = accelstr
	localprevinput["acceltime"] = accelt

	if #accelstr == 1 then
		localprevinput.votestart = leveltime
	end

	if accel then
		localprevinput.accelframes = $+1
	end

	local voteelapsed = leveltime - localprevinput.votestart + 1
	if voteelapsed == localprevinput.accelframes and (voteelapsed == (2 * TICRATE)) then
		print("Accel was pegged to true for 2 seconds. Notify cglitcher unless YOU were holding accel.")
	end

	for ckey,cval in pairs(nvcmds) do
		if cval and cval != localprevinput[ckey] then
			risingEdgeFunctions[ckey](player)
		end
		localprevinput[ckey] = cval
	end
end

-- addHook("PlayerCmd", handlePlayerInput)
nuvote.nva_off = true
nuvote.inputblackout = 50
local function searchPlayerInput()
	-- I think PlayerCmd isn't firing for accel under very specific conditions (unknown why....)
	-- So I'm now checking for playerInput under a ThinkFrame
	nuvote.inputblackout = $-1
	if isdedicatedserver then return end
	if consoleplayer == nil then return end
	if nuvote.nva_off == false then  return end
	if nuvote.inputblackout > 0 then return end
	handlePlayerInput(consoleplayer,consoleplayer.cmd,"ThinkFrame")
end

function nuvote.playVoteMusic(phase)
	if not phase:inTable({"VOTE","VOTEEA", "VOTEEB", "RACENT"}) then
		error(phase.." not a valid vote music to play.")
	end
	nvdbp("Now playing "..phase)
	if not nv_play_vote_music.value then return end
	nuvote.tunes(phase)
end

nuvote.nvchoicewarmup = 3
nuvote.nvchoicemusicstarted = false
function nuvote.checkToPlayMusic()
	if nuvote.nvchoicewarmup > 0 or not nuvote.nva_off or nuvote.nvchoicemusicstarted then
		return
	end
	nuvote.nvchoicemusicstarted = true
	nuvote.playVoteMusic("VOTE")
end

local function fetch_consoleplayer_values()
	if #consoleplayerbus == 0 then return end
	local didvote = false
	for i,kvt in pairs(consoleplayerbus) do
		if kvt["key"] == "nv_choice" and not didvote then
			didvote = true
			COM_BufInsertText(consoleplayer,"vote"..kvt["value"])
			continue
		end
		consoleplayer[kvt["key"]] = kvt["value"]
	end
	consoleplayerbus = {}
end

local function resetChoices()
	for p in players.iterate do
		ui_cursor = choice1
		p.nv_choice = 0
		p.nv_wasafk = 1
	end
	server.nv_result = nil
	server.nv_poolwaspreferred = false
	consoleplayerbus = {}

	if not isserver then return end

	local reset_value = "||||"
	if not isserver then return end
	-- CV_Set(nv_choices,reset_value)
	-- cgs_sendState("nv_choices",reset_value)
	-- nv_choices.value = reset_value
	-- cgs_registers["nv_choices"].value = reset_value
end

function nuvote.resetPlayerState()
	for p in players.iterate do
		p.vetoed = false
	end
end

local superMajorityReq = 5
nuvote.superMinorityReq = 0
local function calcSuperMajorityReq()
	local nonspecsplayers = 0
	for p in players.iterate do
		if not p.spectator then
			nonspecsplayers = $+1
		end
	end
	superMajorityReq = FixedCeil((nv_supermajoritypct.value * nonspecsplayers * FRACUNIT) / 100) / FRACUNIT
	nuvote.superMinorityReq = FixedCeil((nv_superminoritypct.value * nonspecsplayers * FRACUNIT) / 100) / FRACUNIT
	nuvote.superMinorityReq = max($,1)
end

local function resetPlayerInputs()
	-- For some reason players can't input occasionally. Hopefully this works
	for p in players.iterate do
		p.cmd.buttons = 0
	end
	consoleplayerbus = {}
	nuvote.reset_localprevinput()
end
local smr_indicator_sprites = {
	[0] = "SMR1",
	[1] = "SMR2",
	[2] = "SMR3",
}

local colors = {
	[0] = 0,
	[1] = 215
}

local facecolors = {
	[0] = 0,
	[1] = 125
}

local playerIconOffsets = {
	{0,0},{0,23},{0,47},{0,70},
	{20,0},{41,0},{62,0},{83,0},
	{20,70},{41,70},{62,70},{83,70},
	{103,0},{103,23},{103,47},{103,70},
}

local function interpolate(fu_progress,interp_mode)
	if interp_mode == "linear" then return fu_progress end

	if interp_mode == "cosine" then
		local fu_ang1 = FRACUNIT / 180
		local fu_progess_ang = FixedDiv(fu_progress , fu_ang1) / FRACUNIT
		return (-1 * cos( ANG1 * fu_progess_ang)  + FRACUNIT) / 2
	elseif interp_mode == "ease_out" then
		local invp = (FRACUNIT - fu_progress)
		return FRACUNIT - FixedMul(FixedMul(invp,invp),invp)
		-- return min(fu_progress^2,FRACUNIT)
	else

		error("Invalid interpolation mode: "..interp_mode)
	end
end

local function calcAnimPosition(keyframes,curt,v)
	-- prevkf: previous keyframe x, y position
	-- nextkf: next     keyframe x, y position

	local kfidx = nil
	for i,kf in ipairs(keyframes) do
		if curt == kf.t then
			return {x=kf.x,y=kf.y}
		elseif curt < kf.t then
			kfidx = i-1
			-- if kfidx == 0 then kfidx = 1 end -- how to do min in lua??
			break
		end
	end

	-- if we blew past all keyframes, just give the coords of last kf
	if kfidx == nil then
		local kf = keyframes[#keyframes]
		return {x=kf.x,y=kf.y}
	end
	-- v.drawString(0,60, "kfindx="..kfidx.."cur time: "..curt.."kf times "..keyframes[1].t..keyframes[2].t..keyframes[3].t..keyframes[4].t ,V_ALLOWLOWERCASE,"small")


	local prevkf, nextkf = keyframes[kfidx], keyframes[kfidx+1]

	local progress = FixedDiv((curt - prevkf.t) , (nextkf.t - prevkf.t)) 
	local progress_interp = interpolate(progress,nextkf.i)


	local pkf_wgt = FRACUNIT - progress_interp
	local nkf_wgt = progress_interp
	local res = {
		x = (prevkf.x * pkf_wgt + nextkf.x * nkf_wgt) / FRACUNIT,
		y = (prevkf.y * pkf_wgt + nextkf.y * nkf_wgt) / FRACUNIT
	}

	return res
end

local linelen = 16
local actionLines = {}
for i=1,50 do
	actionLines[i] = {
		x = P_RandomRange(0,320 - linelen), -- don't spawn where line can go outside of box
		y = P_RandomRange(0,80 - 1),
		t = P_RandomRange(0,TICRATE)
	}
end

-- this was hell never do this. Just make some sprite elements next time
local function drawActionLine(v,x,y,frame,cyclelen,color)
	local xo = max(0,(frame - 9) * FixedDiv(linelen,7) / FRACUNIT)
	local growth = (FixedDiv(linelen,6) * frame / FRACUNIT)
	local countergrow = ternary(frame-7 < 0,0,growth-linelen)
	local shorten_amt = (FixedDiv(linelen-5,6) * max(frame-9,0)) / FRACUNIT
	-- local shorten = ternary (frame-9 < 0,0,)
	local len = growth - countergrow - shorten_amt
	local xoo = FixedDiv((FixedDiv(linelen,4) / FRACUNIT) * frame, cyclelen) / FRACUNIT
	v.drawFill(x+xo+xoo,y,len,1,color)
	-- v.drawString(x,y,frame,0,"small")
end

local function drawActionLines(v,x,y,bx,by,ct,color)
	local cyclelen = TICRATE
	if #actionLines == 0 then
	end
	for i,al in ipairs(actionLines) do
		drawActionLine(v,x + al.x, y + al.y,al.t,cyclelen,color)
		al.t = (al.t + 1) % cyclelen
		if al.t == 0 then
			-- spawn in a different random location
			al.x =($ + leveltime) % (bx - linelen) 
			al.y =($ + leveltime) % (by - 1) 
		end
	end
end

local function drawVoteStart(v,player,time,endtime)
	-- local keyframes = {0,20/100 * endtime, 80/100 * endtime,endtime}
	local t_vt_enter = 0
	local t_vt_main = FixedDiv(10 , 100) * endtime  / FRACUNIT
	local t_vt_exiting = FixedDiv(90 , 100) * endtime  / FRACUNIT
	local t_vt_exit = endtime
	local votingtimekf = {
		{x=0,y=100,t=t_vt_enter,i='linear'},
		{x=320 / 2 - 15,y=100,t=t_vt_main,i='linear'},
		{x=320 / 2 + 15,y=100,t=t_vt_exiting,i='linear'},
		{x=320,y=100,t=t_vt_exit,i='cosine'}
	}

	local barkf = {
		{x=-320,y=65,t=t_vt_enter,i='linear'},
		{x=0,y=65,t=t_vt_main,i='cosine'},
		{x=0,y=65,t=t_vt_exiting,i='linear'},
		{x=320,y=65,t=t_vt_exit,i='cosine'}
	}

	v.fadeScreen(0xFF00, 16)
	local vt_pos = calcAnimPosition(votingtimekf, time,v)
	local bar_pos = calcAnimPosition(barkf,time,v)
	v.drawFill(bar_pos.x,bar_pos.y,320,80,243)
	drawActionLines(v,bar_pos.x,bar_pos.y,320,80,50,210) -- match the bar exactly
	v.drawString(vt_pos.x,vt_pos.y, "Voting Time!!",V_ALLOWLOWERCASE,"center")
end

local nvvotetimer = 20 * TICRATE
local nvshowresulttimer = (4 * TICRATE) + 5 -- length of VOTEEB


local function drawMapChoice(v,player,x,y,choiceId,mapid,isEncore)
    local mapName = G_BuildMapName(mapid)
    local patchName = mapName.."P"
	local mapp
	if choiceId == 4 and not (nv_state == NV_RESULT and server.nv_result and (server.nv_result >= 4)) then
		-- show random, reveal if it was chosen
		mapp = v.cachePatch("RANDOMLV")
	else
		mapp = v.patchExists(patchName) and v.cachePatch(patchName) or v.cachePatch("BLANKLVL")
	end

    -- facerank
	
	local numvoters,numvotedforcurrent = 0, 0
	for i,pio in ipairs(playerIconOffsets) do
		local p
		if huddebug then
			p = players[1]
		else
			p = players[i-1]
		end

		if (p == nil or p.spectator) then continue end
		numvoters = $+1
		if p.nv_choice != choiceId and not huddebug then continue end
		numvotedforcurrent = $+1
		local px, py = x + pio[1] , y + pio[2]
		local skin = skins[p.mo.skin]
		local pcolor = v.getColormap(skin.name,p.skincolor)

		local facerank = skin and v.cachePatch(skin.facerank) or v.cachePatch("M_NORANK")

		if p == player then
			-- show the player where they are
			v.drawFill(px-1,py-1,18,18, FixedMul(sin(leveltime * 70/100 * ANG10),5) + 126)
		end
		v.draw(px,py, facerank, 0, pcolor) 
	end

	local isInducedHellchoice = (choiceId == 4 and server.nv_result == 5)
	if nv_state == NV_RESULT and ((server.nv_result == choiceId) or isInducedHellchoice) then
		v.drawFill(x + 15,y + 15,89,56,colors[leveltime / 4 % 2])
	end

	if nv_state == NV_CHOICE and not player.spectator and consoleplayer.nv_choice == 0 and string.sub(ui_cursor.id, -1) == tostring(choiceId) then
		v.drawFill(x + 15,y + 15,89,56,colors[leveltime / 16 % 2])
	end



	-- map patch

    local scale = FRACUNIT - (1000 * scalar)
	

	if v.patchExists(smr_indicator_sprites[leveltime / 8 % 3]) then
		local smr_ind = v.cachePatch(smr_indicator_sprites[leveltime / 8 % 3])
		if smr_ind != nil then
			v.drawScaled(curx * FRACUNIT,cury * FRACUNIT,scale,smr_ind,V_40TRANS)
		end
	end

    -- v.drawScaled((x + 16) * FRACUNIT,(y + 16) * FRACUNIT,scale,mapp,V_INVERTMAP)
    v.drawScaled((x + 16) * FRACUNIT,(y + 16) * FRACUNIT,scale,mapp,0)


	-- insert hell indicator patch here
	if mapIsHell(mapid) and not (choiceId == 4 and nv_state != NV_RESULT) then
		v.drawString(x + 102,y + 17,"hell",V_50TRANS,"small-right")
	end
	-- v.drawString(x + 94,y + 17,choiceId,V_50TRANS)

	local choiceTtl = "random"
	if (choiceId != 4 or (nv_state == NV_RESULT and server.nv_result and server.nv_result >= 4)) then
		local zonetitle = mapheaderinfo[mapid].zonttl
		zonetitle = ternary(string.lower(zonetitle) == "zone","",zonetitle)
		local actnum = mapheaderinfo[mapid].actnum
		actnum = ternary(actnum == "",""," "..actnum)
		choiceTtl = mapheaderinfo[mapid].lvlttl.." "..zonetitle..actnum
	end
	v.drawString(x+101,y+60,choiceTtl,0 ,"thin-right")

	-- majority indication
	if nv_state == NV_CHOICE and numvoters > 4 then
		local pct = 80
		local colors = nil
		local almost =  FixedCeil((pct * superMajorityReq * FRACUNIT) / 100) / FRACUNIT
		if numvotedforcurrent >= (superMajorityReq) then
			colors = {212, 213, 214, 215, 216}
			drawMarqueeRect(v,x + 17,y + 17,85,54,colors,2)
		elseif numvotedforcurrent >= almost
			colors = {1, 208, 209, 210, 211}
			drawMarqueeRect(v,x + 17,y + 17,85,54,colors,20)
		end
	end

	if not isEncore then return end

	local bob = sin(leveltime * 95/100 * ANG10)
	v.drawScaled(
		(x + 60) * FRACUNIT,
		bob * 5/2 + (y + 40) * FRACUNIT,
		FRACUNIT * 70/100,
		v.cachePatch("RUBYICON"),
		0
	)
end

local function drawNextMap(v,x,y,pctscale,mapid,isEncore)
	local scale = (pctscale * FRACUNIT) / 100
    local mapName = G_BuildMapName(mapid)
    local patchName = mapName.."P"
	local mapp = v.patchExists(patchName) and v.cachePatch(patchName) or v.cachePatch("BLANKLVL")

	local width, height = (scale * 162) / FRACUNIT, (scale * 102) / FRACUNIT
	if pctscale < 50 then
		width = $+2
		height = $+2
	elseif pctscale < 100 then
		width = $+1
		height = $+1
	end
	v.drawFill(x,y,width,height,colors[leveltime / 16 % 2])

	-- map patch

    v.drawScaled((x + 1) * FRACUNIT,(y + 1) * FRACUNIT,scale,mapp,0)

	-- build title
	local choiceTtl = ""
	local zonetitle = mapheaderinfo[mapid].zonttl
	zonetitle = ternary(string.lower(zonetitle) == "zone","",zonetitle)
	local actnum = mapheaderinfo[mapid].actnum
	actnum = ternary(actnum == "",""," "..actnum)
	choiceTtl = mapheaderinfo[mapid].lvlttl.." "..zonetitle..actnum


	v.drawString(x + width,y + height-10,choiceTtl,0 ,"thin-right")

	if not isEncore then return end

	local bob = sin(leveltime * 95/100 * ANG10)
	v.drawScaled(
		(x + 60) * FRACUNIT,
		bob * 5/2 + (y + 40) * FRACUNIT,
		FRACUNIT * 70/100,
		v.cachePatch("RUBYICON"),
		0
	)
end

nuvote.mapChoice_offsets = {
	{13,17}, {154,17},
	{33,110}, {174,110}
}

local function drawMapChoices(v,player, mapChoices)
	local coffset = 10

	for choiceID, mapChoice in ipairs(mapChoices) do
		local mapid,isEncore = mapChoice[1], mapChoice[2]
		local cx,cy = nuvote.mapChoice_offsets[choiceID][1], nuvote.mapChoice_offsets[choiceID][2]
		-- local xoffset = coffset *  (1 - (2 * (choiceID % 2))) -- add or subtract 10, switching at each index
		-- cx = cx + xoffset

		-- if result is a random induced hell, send the hell map details
		if choiceID == 4 and server.nv_result == 5 then
			mapid,isEncore = mapChoices[5][1], mapChoices[5][2]
		end

		drawMapChoice(v,player,cx,cy,choiceID,mapid,isEncore)

		if choiceID == 4 then break end

	end
end

local sin = sin

-- local nv = {}

local function drawDebug(v)
	if not CV_FindVar("nv_debug").value then return end
    v.drawFill(curx,cury,1,1,colors[leveltime / 8 % 2])
    v.drawNum(280,180,curx)
    v.drawNum(310,180,cury)

	-- v.drawString(300, 120,"INTERSCREEN: "..v.width().." "..v.height().." "..ternary(interscreen==nil,"",interscreen),0,"small-right")

	local curs = "curs: "
	for i=1,9 do
		curs = $..i..":"..nuvote["cur"..i].." "
	end

	v.drawString(310,160,curs,0,"small-right")

	local accels = "accels: "
	for i,sig in ipairs(localprevinput["accelprev"]) do
		accels = $.." "..ternary(sig,"1","0").."  "
	end

	local accelt = "time  : "
	for i,t in ipairs(localprevinput["acceltime"]) do
		accelt = $..t.." "
	end

	v.drawString(310,130,tostring(consoleplayer.kartstuff[k_sneakertimer]),0,"small-right")

	v.drawString(310,140,accelt,0,"small-right")
	v.drawString(310,150,accels,0,"small-right")

	v.drawString(0, 105,"playerstate: "..(consoleplayer.playerstate & PST_DEAD).."Timeover cam: "..consoleplayer.kartstuff[k_timeovercam].."TIMEOVER: "..(consoleplayer.pflags & PF_TIMEOVER),0,"small")

	-- local dbmsg = string.format("x: %d y: %d s: %d wh: %s uw: %s (%s)",nv.x,nv.y,nv.s,nv.wh, tostring(nv.uw),v.renderer())
	-- v.drawString(0,95,dbmsg,0,"small")

	local ins = ""
	for _,k in ipairs({"up","down","left","right","accel"}) do
		local vv = localprevinput[k]
		ins = $..k..": "..ternary(vv,1,0).." "
	end
	ins = $.."Buttons "..consoleplayer.cmd.buttons.." "
	ins = $.."Side "..consoleplayer.cmd.driftturn.." "
	v.drawString(0,115,ins,0,"small")

	v.drawString(0,125,"Cursor: "..ui_cursor.id,0 ,"small")


	-- current time
	v.drawString(0,135,"Time: "..cgt_fulltime_readout(cg_get_current_time(),"-5:00").." (EDT)",0 ,"small")

	-- state
	v.drawString(0,155,"states: P="..prevState.." C="..nv_state,0 ,"small")

	local css = "choices:"
	local nvchoices = get_nv_choices()
	if nvchoices == nil then
		nvchoices = {{"X",false},{"X",false},{"X",false},{"X",false},{"X",false}}
	end
	for i,c in ipairs(nvchoices) do
		css = $.." "..c[1]
	end

	v.drawString(0, 145,css,0,"small")

	-- player choices
	local cs = "selections: "-- 1 2 3 4 5 6 7 8 9 10 11 12 13 "
	local as = "was        afk: "-- 1 2 3 4 5 6 7 8 9 10 11 12 13 "
	local ss = "scores:          "
	local ps = "positions: "
	for p in players.iterate do
		local pcs = "X"
		local pwa = "X"
		if p.spectator then pcs,pwa = "S","S"
		elseif p.nv_choice == nil then pcs = "N"
		elseif p.nv_choice == 0 then pcs = "?"
		elseif p.nv_choice != nil then  pcs = nvchoices[p.nv_choice][1] end

		if p.nv_wasafk == nil then pwa = "N"
		else pwa = p.nv_wasafk
		end

		cs = $.." "..pcs
		as = $.." "..pwa
		ss = $.." "..p.score
		ps = $.." "..p.kartstuff[k_position]
	end
	v.drawString(0,165,cs,0,"small")
	v.drawString(0,175,as,0,"small")
	-- v.drawString(0,185,ss,0,"small")
	v.drawString(0,185,ps,0,"small")


	local rms = "Recent maps: "
	for m in QueuedKeys.iterate(nv_recent_maps) do
		rms = $..m.." "
	end
	-- v.drawString(0,195,rms,0,"small")

	-- local mcolors = {216, 215, 214, 213, 212, 211, 210, 209, 208, 1}
	-- local mcolors = {1, 208, 209, 210, 211, 212, 213, 214, 215, 216}

	-- drawMarqueeRect(v,curx,cury,50,50,mcolors,20)

end

local function noop(v,player)
    return
end

local function getReactionString(mapid)
	if mapIsHell(mapid) then return "uh oh..." end
	mapid = tonumber(mapid)
	-- nvdbp("mapid",mapid,"access",mapReactions[mapid])
	if mapReactions[mapid] != nil then
		return mapReactions[mapid]
	end
	return ""
end

local function handleNewPlayer(p)
	-- reset function
	p.nv_wasafk = 1
	p.nv_choice = 0
	p.vetoed = false
end

local function checkForNewPlayers()
	for p in players.iterate do
		if p == nil then nvdbp("Hit nil player") end
		if p.nv_wasafk == nil then
			handleNewPlayer(p)
		end
	end
end

local randominducedhell = true
local wasmajorityforced = false
local calculatedresult = false
local function calculateResult()
	local votingPlayers = 0
	local choiceTallys = { 0,0,0,0 }
	local choices = {}
	local winner = nil
	randominducedhell = true

	-- tally votes

	for p in players.iterate do
		if p.spectator then continue end
		votingPlayers = $ + 1
		choices[#choices + 1] = p.nv_choice
		if p.nv_choice == 0 then 
			p.nv_choice = 4 -- default to random if player didn't vote
		end

		if p.nv_choice ~= 4 then
			randominducedhell = false
		end

		choiceTallys[p.nv_choice] = $ + 1
	end

	-- determine winner

	if votingPlayers == 1 or not nv_random_induces_hell.value then
		randominducedhell = false
	end
	--first check if we're going to hell
	if randominducedhell then
		winner = 5
		wasmajorityforced = true
	-- next check if there were any supermajority winners
	elseif votingPlayers != 0 then
		for i, ct in ipairs(choiceTallys) do
			if ct >= superMajorityReq then
				winner = i
				-- don't show roulette
				wasmajorityforced = true
				break
			end
		end
	else
	-- in hardcode, if there are no players present during vote, it goes to
	-- NextLevel in the current map's soc. But hardly any mappers define this 
	-- and it always goes to map01. So instead I just choose one already listed at random.
		winner = ((leveltime % 4) + 1)
		wasmajorityforced = true -- no roulette
	end

	-- there was no super majority
	local windex = nil
	if winner == nil then
		nvdbp("Supermajority was NOT reached")
		-- keep track of any choices that pass the min vote requirement
		-- If none pass, choose the choice(s) with most votes. There could be more than one
		-- only play the roulette if more than 1 choice is at stake
		local eligibles = {}
		local mostvotes = {choiceid={},num_votes=0}
		for i, ct in ipairs(choiceTallys) do
			nvdbp("tally idx",i,"ct",ct,"mostvotes num",mostvotes.num_votes)
			if ct > mostvotes.num_votes then
				mostvotes.choiceid = {i}
				mostvotes.num_votes = ct
			elseif ct == mostvotes.num_votes then
				table.insert(mostvotes.choiceid,i)
			end
			if ct < nuvote.superMinorityReq then continue end
			nvdbp("supMinReq",nuvote.superMinorityReq,"ct",ct)
			table.insert(eligibles,i)
		end

		nvdbp("mostvotes",table.concat(mostvotes.choiceid,","),"ct",mostvotes.num_votes)
		nvdbp("eligibles",table.concat(eligibles,","))

		if #eligibles == 1 then
			-- only this one passed superminority minimum
			winner = eligibles[1]
			wasmajorityforced = true
			nvdbp("only 1 elibile")
		elseif #eligibles == 0 then
			-- No choice passed superminority minimum, so let's choose the
			-- map with most votes (there might be a tie)
			nvdbp("No eligibles")
			if #mostvotes.choiceid == 1 then
				nvdbp("There was just one map with most votes")
				wasmajorityforced = true
			end
			windex = (leveltime % #mostvotes.choiceid) + 1
			winner = mostvotes.choiceid[windex]
		else
			-- multiple choices are eligible for routlette
			nvdbp("There are multiple eligibles")
			windex = (leveltime % #eligibles) + 1
			winner = eligibles[windex]
		end
	end

	
	if winner == nil then
		print("Bug in tally logic resulted in nil winner. Selecting Mickey Mouse.")
		local msg = "Debug info: Votes: "..table.concat(choices,",").." SuperMajPct: "..nv_supermajoritypct.value.." SuperMajReq: "..superMajorityReq.." SuperMinPct: "..nv_superminoritypct.value.." SuperMinReq: "..nuvote.superMinorityReq
		print(msg)
		if #choices != 0 then
			windex = (leveltime % #choices) + 1
		else
			print("Aiieee!! No votes either!? I'm choosing random; this shit is ridiculous..")
			windex = (leveltime % 4) + 1
		end
		winner = choices[windex]
	end
	
	nvdbp("Winner",winner)
	server.nv_result = winner
	calculatedresult = true
end

local function addResultToRecentMaps()
	local nextmapid = get_nv_choices()[server.nv_result][1]
	QueuedKeys.add(nv_recent_maps,nextmapid)
end

local lastplayedsound = 0
local function playResultSounds()
	-- sfx_kc39 SRB2Kart: Voting screen roulette sound
	-- sfx_kc4a SRB2Kart: Choosing a map on the voting screen
	-- sfx_kc48 SRB2Kart: A map being selected on the voting screen.
	-- sfx_s3k68 Final lap
	-- sfx_noooo1 this is bad
	-- sfx_noooo2 gasp
	-- sfx_yeeeah yeeaaah!
	if isdedicatedserver then return end

	-- DO NOT PLAY ON EVERY TIC OR NO ONE WILL TRUST THIS MOD EVER!!!!
	-- this is a safety net in case erroneous operation calls this on every tic
	local currenttime = cg_get_current_time()
	local diff = currenttime - lastplayedsound
	if diff < 5 then return end
	lastplayedsound = currenttime

	NV_playsound(nil, sfx_kc48)
	local selectedmapid = get_nv_choices()[server.nv_result][1]
	local selectionishell = mapIsHell(selectedmapid)
	if selectionishell and not randominducedhell then
		NV_playsound(nil, sfx_noooo1) -- this is bad!!
		-- sidenote: is this shulk xenoblade??
	elseif selectionishell and randominducedhell then
		NV_playsound(nil, sfx_noooo2)
	elseif consoleplayer.nv_choice == server.nv_result then
		NV_playsound(nil, sfx_yeeeah,consoleplayer)
	end
end

function nuvote.cast_mapwarp(mapid,isencore)
	nvdbp("casing warp")
	if nuvote.mapwarpcasted then return end
	local encorepart = ternary(isencore," -encore","")
	local command = "map "..mapid..encorepart
	COM_BufInsertText(server,command)
	nuvote.mapwarpcasted = true
end

function nuvote.randommap()
	local nextinfo = nuvote.randommap_helper("randommap")
	local mapid, encore = nextinfo[1], nextinfo[2]
	nuvote.cast_mapwarp(mapid, encore)
end

local function enterLevel()
	if nuvote.mapwarpcasted then return end
	local result = get_nv_choices()[server.nv_result]
	local mapstring = result[1]
	local isencore = result[2]
	nvdbp("casting warp from enterlevel")
	nuvote.cast_mapwarp(mapstring,isencore)
end

local function checkIfAllVoted()
	local nonspecExists = false
	for p in players.iterate do
		if not p.spectator then nonspecExists = true end
		if p.nv_choice == nil then p.nv_choice = 0 end
		if not p.spectator and p.nv_choice == 0 then return end
	end

	-- prevent from voting ending if all players in the room are spec
	if not nonspecExists then return end

	-- If we reach here, then all players have voted, let's transition
	calculateResult()
	-- nvdbp(wasmajorityforced)
	if wasmajorityforced then
		nv_state = NV_RESULT
	else
		nv_state = NV_ROULETTE
	end
end

local function allowTeamChangePlaying()
	-- During voting, allow a player to come back in to vote
	for p in players.iterate do
		if not p.spectator then continue end
		if p.pflags & PF_WANTSTOJOIN then
			p.spectator = 0
			restoreKartPosition(p)
			if p.mo then 
				p.mo.subsector.sector.special = 0
				p.mo.friction = 0
			end
			nuvote.reset_localprevinput()
		end
	end
end

local function countdownVote()
	nvvotetimer = $-1
	--freemin
	K_SetExitCountdown(TICRATE*6000) -- Just keep setting this to ensure hardcode doesn't exit us 
	if server.hmfinishtimer then server.hmfinishtimer = 5000 end
	if server.kvrfinishtimer then server.kvrfinishtimer = 5000 end
	if nvvotetimer == 0 then
		-- we're transitioning. Calc before to not get 1 frame of roulette if majority
		calculateResult()
		if wasmajorityforced then
			nv_state = NV_RESULT
		else
			nv_state = NV_ROULETTE
		end
	end
end

local function countdownVoteCooldown()
	if nv_state != NV_CHOICE then return end
	for p in players.iterate do
		if p.nv_votecooldown == nil then p.nv_votecooldown = 0 end
		p.nv_votecooldown =$-1
	end
end

local function countdownShowResult()
	nvshowresulttimer = $-1
	K_SetExitCountdown(TICRATE*6000) -- Just keep setting this to ensure hardcode doesn't exit us 
	if server.hmfinishtimer then server.hmfinishtimer = 5000 end
	if server.kvrfinishtimer then server.kvrfinishtimer = 5000 end

	if nvshowresulttimer == 0 then
		nv_state = NV_WAIT
		-- K_SetExitCountdown(0)
		-- G_ExitLevel()
	end
end

local function resetVoteTimers()
	-- resets all timers related to voting
	nvvotetimer = CV_FindVar("votetime").value * TICRATE
	nvshowresulttimer = (4 * TICRATE) + 5 -- length of VOTEEB
	for p in players.iterate do
		p.nv_votecooldown = 0
	end
end

function nuvote.playNVNEXTSounds_helper()
	S_StartSound(nil,sfx_itemup)
end

function nuvote.playNVNEXTSounds()
	local nummaps = 0
	if cq_hasmaps != nil then
		nummaps = #cq_getallmaps()
	elseif aTourney != nil then
		nummaps = #aTourney.tourneyMapList
	end

	local noop = function() return end
	for i=1,nummaps do
		local batch = {{nuvote.playNVNEXTSounds_helper},{noop},{noop}}
		funcQueue.add_callback_batch(batch)
	end
end

function nuvote.disableSneakerPanels()
	for p in players.iterate do
		if not p or not p.valid then continue end
		p.kartstuff[k_sneakertimer] = 0
		p.kartstuff[k_spinouttimer] = 0
		p.kartstuff[k_squishedtimer] = 0
		if p.mo then
			-- for boostpads, or lua triggers
			p.mo.subsector.sector.special = 0
			p.mo.friction = 0
		end
	end
end

function nuvote.disablePlayerControl()
	local nocontrolticks = nvvotetimer + nvshowresulttimer
	-- also disable any effects which might influence p.cmd
	for p in players.iterate
		if p and p.valid then
			p.kartstuff[k_sneakertimer] = 0
			p.kartstuff[k_spinouttimer] = 0
			p.kartstuff[k_squishedtimer] = 0
			p.exiting = 8
			p.powers[pw_nocontrol] = nocontrolticks
			if p.mo then
				-- for boostpads, or lua triggers
				p.mo.subsector.sector.special = 0
				p.mo.friction = 0
			end
		end
	end
end

local hm_spec_value
local function pacifyHMSoftlockDetection()
	-- pacify hostmod softlock detection for now
	if not isserver then return end
	local hm_spec = CV_FindVar("hm_specbomb")
	if hm_spec == nil then return end
	hm_spec_value = hm_spec.value
	CV_Set(hm_spec,"Off")
	if tonumber(HOSTMOD) > 16 then
		CV_Set(CV_FindVar("hm_specbomb_antisoftlock"),"Off")
	end
end

local function restoreHMSoftlockDetection()
	if not isserver then return end
	if hm_spec_value == nil then return end
	local hm_spec = CV_FindVar("hm_specbomb")
	CV_Set(hm_spec,hm_spec_value)
end

function nuvote.checkVeto()
	-- new veto system is great, but it added more work for me lmao
	-- though now we don't know for sure who is saying no to veto-ing
	-- assumers we're running in nv_state == NV_DEFAULT
	if HOSTMOD == nil then return end
	if tonumber(HOSTMOD) <= 16 then return end
	if server.HMvetostate != 1 then return end -- 1 = VS_VOTING

	for p in players.iterate do
		if p.vetoed != nil and p.vetoed then continue end
		if p.HMvetotrigger != nil and p.HMvetotrigger >= TICRATE/2 then
			p.vetoed = true
			local isinitiator = true
			for p2 in players.iterate do
				if p2 == p then continue end
				if p2.vetoed != nil and p2.vetoed then
					isinitiator = false
					break
				end
			end
			-- check if they're the first
			NV_recordRtv(p,true,isinitiator)
		end
	end
end

function nuvote.advanceMatchNum()
	if leveltime and leveltime == 4 then
		nuvote.match_num = $ + 1
	end
end

local function pollMatchEnded()
	if nv_state != NV_DEFAULT then return end

	local matchEnded = false

	-- First check to see if one of the custom intermissions is running
	-- Kart VS/VR does the same thing as NVJR but with different names

	local customintkeys = {
		{
			finishstate = "hmfinishstate",
			finishtimer = "hmfinishtimer",
			trigger = "HM_ResultsAutoTrigger"
		},
		{
			finishstate = "kvrfinishstate",
			finishtimer = "kvrfinishtimer",
			trigger = "KVR_ResultsAutoTrigger"
		},
	}

	local customintloaded = false

	for _, customint in ipairs(customintkeys) do
		if server[customint.finishstate] == nil continue end
		customintloaded = true

		if not server[customint.finishstate] then continue end -- race still ongoing
		if server[customint.finishtimer] < 5 then
			matchEnded = true
			server[customint.finishtimer] = 4
			rawset(_G,customint.trigger,false)
			server[customint.finishstate] = false
			nvdbp("Trigger",HM_ResultsAutoTrigger)
		end
	end

	
	if not customintloaded then
		-- This code based off of freeman and Snu
		-- count # of finished players including dnf
		local pdone, ptotal = 0, 0
		for p in players.iterate
			if not p.spectator then ptotal = $+1 end
			if p.exiting and p.exiting < 8 then p.exiting = 8 end
			if (p.exiting and p.exiting == 8) or (p.pflags&PF_TIMEOVER) then
				if not p.spectator then pdone = $+1 end
			end
		end

		-- match ended when all non-spec players are finished
		if not ptotal then return end
		if (pdone >= ptotal or exitcountdown == 8) and not server.hmfinishstate then
			K_SetExitCountdown(TICRATE*6000)
			matchEnded = true
		end
	end

	if matchEnded then
		-- We completed a race here! Now put it in recent maps if not already
		nuvote.disablePlayerControl()
		if not QueuedKeys.contains(nv_recent_maps,gamemap) then
			QueuedKeys.add(nv_recent_maps,gamemap)
		end

		if (CV_FindVar("advancemap").string != "Vote") then
			nv_state = NV_WAIT
			-- print("advancemap wasn't vote ".. CV_FindVar("advancemap").string)
			-- G_ExitLevel()
		elseif maps_are_queued() then
			-- currently map queues only work if advancemap is on "Vote"
			nv_state = NV_NEXT
		else
			nv_state = NV_CHOICE
		end
	end
end

local function exitlevel2(p)
	if not nv_enableNV.value then
		CONS_Printf(p,"exitlevel2 only works when NuVote is enabled (cvar=nv_enableNV)")
		return
	end

	-- only do stuff during normal gameplay
	if (nv_state & (NV_CHOICE|NV_ROULETTE|NV_RESULT|NV_NEXT)) then
		print("We've already exited! Try again later.")
		return
	elseif nv_state == NV_WAIT then
		print("Queing up exitlevel2 at the start of the next race...")
		nuvote.exitlevel2_queued = true
		return
	end

	-- apply DNF to any non-finished players
	for p in players.iterate do
		if p.spectator or p.exiting continue end
		p.exiting = TICRATE * 60
		p.pflags =$|PF_TIMEOVER
		p.lives = 0
		p.playerstate = PST_DEAD
	end

	for p in players.iterate do
		COM_BufInsertText(p, "tunes KRFAIL")
	end

	if maps_are_queued() then
		nv_state = NV_NEXT
	elseif false
		print("hi")
		-- Logic to show kart vs, maybe
	else
		nv_state = NV_CHOICE
	end
end
COM_AddCommand("exitlevel2",exitlevel2,COM_ADMIN)

local function exitlevel_override(p,force_flag)
	if not nv_enableNV.value then
		G_ExitLevel()
		return
	end
	
	if force_flag != "-force" then
		CONS_Printf(p,"warning: exitlevel will skip over NuVote. Use exitlevel2 to exit to NuVote. Use 'exitlevel -force' to use true exitlevel")
		return
	end
	
	G_ExitLevel()
end

-- I might get crucified for this. Willing to remove, but will prevent erroneous operation
COM_AddCommand("exitlevel",exitlevel_override,COM_ADMIN)

nuvote.exitlevel2_queued = false
function nuvote.checkForExitlevel2()
	if not nuvote.exitlevel2_queued then return end
	nuvote.exitlevel2_queued = false
	COM_BufInsertText(server, "exitlevel2")
end


-- override Hostmod behavior so that exitlevel doesn't get called
addHook("PlayerMsg", function(source, msgtype, target, msg)
	-- don't do anything if hostmod not loaded
	if server.HMvotable == nil then return false end 
	if msgtype != 0 then return end

	
	local HMhasVeto = tonumber(HOSTMOD) > 16
	local rtvvote = server.HMvtimer and server.HMvcommand:inTable({"exitlevel", "exitlevel2"})
	if msg == "rtv" then 
		-- record that someone rtv'd a map
		NV_recordRtv(source,true,true)
		source.vetoed = true
		
		if HMhasVeto then return false end
		if server.HMvtimer then
			COM_BufAddText(source, "voteyes")
			return true
		else
			COM_BufAddText(source, "vote exitlevel2")
			return true
		end
	elseif msg == "y" and rtvvote then
		NV_recordRtv(source,true,false)
		return true
	elseif msg == "n" and rtvvote then
		NV_recordRtv(source,false,false)
		return true
	end

	return false
end)

local function resetState()
	nuvote.mapwarpcasted = false
	if HM_ResultsAutoTrigger != nil then HM_ResultsAutoTrigger = true end
	if KVR_ResultsAutoTrigger != nil then KVR_ResultsAutoTrigger = true end
	wasmajorityforced = false
	randominducedhell = true
	calculatedresult = false
	nuvote.mapwasqueued = 0
	nuvote.nuvote_preempted = false
	nuvote.intermission_countdown = 0
	nuvote.nvchoicemusicstarted = false
	nuvote.nvchoicewarmup = 3
	nv_animation = nil
	nuvote.prev_preferred_every = (CV_FindVar("nv_preferred_every").value)
	nuvote.resetPlayerState()
end

nuvote.isoffsets = {}
nuvote.isoffsets["19201080"]={x=-32,y=-20,s=13100}
nuvote.isoffsets["16801050"]={x=-8,y=-5,s=3275}
nuvote.isoffsets["16001200"]={x=-32,y=-20,s=13100}
nuvote.isoffsets["1600900"]={x=-66,y=-12,s=8515}
nuvote.isoffsets["1366768"]={x=-68,y=-42,s=23580}

nuvote.isoffsets["1440900"]={x=-20,y=-13,s=8515}
nuvote.isoffsets["12801024"]={x=-45,y=-28,s=18340}
nuvote.isoffsets["1280960"]={x=-32,y=-20,s=13100}
nuvote.isoffsets["1280720"]={x=-54,y=-34,s=22270}
nuvote.isoffsets["1152864"]={x=-70,y=-44,s=28820}

nuvote.isoffsets["1024768"]={x=-45,y=-28,s=18340}
nuvote.isoffsets["800600"]={x=-120,y=-50,s=32750} -- why is wide used here??
nuvote.isoffsets["640480"]={x=-32,y=-20,s=13100}
nuvote.isoffsets["320240"]={x=-32,y=-20,s=13100}

function nuvote.drawVoteBg(v,p)
	if not nv_show_vote_bg.value then return end
	-- Race bg: INTERSC
	-- Battle bg: BATTLSC
	-- Small: R
	-- Wide: W
	-- screen res = 320 x 200

	-- first calc which patch to use
	-- copied logic from y_inter.c
	local usewide = ternary(((v.width() / v.dupx()) == 400) and v.renderer() == "software",1,0)
	local isbattle = ternary(gametype == 1,1,0)
	local pidx = tostring(usewide)..tostring(isbattle)
	local plu = {}
	plu["00"] = "INTERSCR"
	plu["01"] = "BATTLSCR"
	plu["10"] = "INTERSCW"
	plu["11"] = "BATTLSCW"
	local patchname = plu[pidx]
	local bg_patch = v.cachePatch(patchname)
	
	-- Calculate offsets
	local wh = tostring(v.width())..tostring(v.height())
	local isoffsets = {x=0,y=0,s=0}
	if v.renderer() == "software" then
		isoffsets = nuvote.isoffsets[wh] or {x=0,y=0,s=0}
	end

	local xoffset = (0+curx+isoffsets.x)
	local yoffset = (0+cury+isoffsets.y)
	local scale = (FRACUNIT + isoffsets.s + nuvote.cur1)

	-- nv.x = xoffset
	-- nv.y = yoffset
	-- nv.s = nuvote.cur1
	-- nv.wh = wh
	-- nv.uw = usewide

	v.drawScaled(xoffset * FRACUNIT,yoffset * FRACUNIT,scale,bg_patch,0)

end

local function drawNVCHOICE(v, player)
	if nuvote.nvchoicewarmup > 0 then return end
	v.fadeScreen(0xFF00, 16)
	nuvote.drawVoteBg(v,player)


	-- adding space for other elements, now that we can vote with controller
	-- local votePrompt = "vote in chat with \1301 \128/ \1302 \128/ \1303 \128/ \1304"
	local votePrompt = "This round's choices"
	v.drawString(160,4, votePrompt,V_SNAPTOTOP|V_ALLOWLOWERCASE,"center")

	if showSMR.value then
		v.drawString(313,18, "For majority: "..superMajorityReq,V_SNAPTORIGHT,"small-right")
		v.drawString(313,24, "For eligibility: "..nuvote.superMinorityReq,V_SNAPTORIGHT,"small-right")
	end

	-- draw the four choices
	local nvchoices =get_nv_choices()
	if nvchoices != nil then
		drawMapChoices(v,player,nvchoices)
	end

	local voteTimeString = "vote ends in "..string.format("%02i",nvvotetimer/TICRATE)

	-- thank you freeman and Snu
	-- store highlight colour here for use in the hud throughout
	local hilicol
	if (CV_FindVar("menuhighlight").value)
		hilicol = CV_FindVar("menuhighlight").value
	elseif (modeattacking)
		hilicol = V_ORANGEMAP
	else
		hilicol = ((gametype == GT_RACE) and V_SKYMAP or V_REDMAP)
	end
	v.drawString(159,189,voteTimeString, V_SNAPTOBOTTOM|hilicol,"center")


	v.drawString(101,13,"Type \130!nvhelp\128 in chat for more info", V_SNAPTOTOP|V_ALLOWLOWERCASE,"small")

	nuvote.drawClock(v,player)
	-- debug
    drawDebug(v)
end

nuvote.timehint_offset = 0
-- nuvote.timehint = "chat !nvclockhelp"
nuvote.timehint = "chat !nvclockhelp to setup.  "
nuvote.hintchars = 17
function nuvote.drawClock(v,player)
	if not CV_FindVar("nv_enableTime").value then return end
	-- If server didn't set up utc then just don't show it
	local utc = cg_get_current_time()
	if utc < 1698020496 then return end

	local showtime, tzoffset, ampm = false, 0, false
	local showhint = ""
	-- nvdbpo("drawclock",tostring(player.nvsettings))

	if player.nvsettings != nil then
		showtime = nvs_get(player.nvsettings,"nv_showtime")
		tzoffset = nvs_get(player.nvsettings,"nv_tzoffset")
		ampm = nvs_get(player.nvsettings,"nv_ampm")
		local clocknoob = (
			nvs_notset(player.nvsettings,"nv_showtime") and
			nvs_notset(player.nvsettings,"nv_tzoffset") and
			nvs_notset(player.nvsettings,"nv_ampm")
		)
		-- choose characters for scrolling help
		local showstr = ""
		for i=1,nuvote.hintchars do
			local j = (((i - 1) + nuvote.timehint_offset) % #nuvote.timehint) + 1
			local nextchar = nuvote.timehint:sub(j,j)
			nextchar = ternary(nextchar == " ","  ",nextchar)
			nextchar = ternary(nextchar == ".",". ",nextchar)
			nextchar = ternary(nextchar == "!"," !",nextchar)
			showstr = $..nextchar
		end
		showhint = ternary(clocknoob,"("..showstr..")","")
	end
		
	if showtime then
		local timestr = cgt_condensed_readout(cg_get_current_time(),tzoffset,ampm)..showhint
		v.drawString(0,196,timestr, V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_30TRANS|V_ALLOWLOWERCASE,"small")
	end
end

addHook("ThinkFrame",function()
	if not consoleplayer then return end
	consoleplayer.nvtimeingame = $ or 0
	consoleplayer.nvtimeingame = $ + 1
	
	-- don't scroll while player is in game
	local playerisdone = (consoleplayer.exiting and consoleplayer.exiting != 0) or (consoleplayer.pflags&PF_TIMEOVER) 
	-- nvdbp("pdone",consoleplayer.exiting,"spec",consoleplayer.spectator,"in game",nuvote.prev_gamestate == nuvote.GS_GAME,"timeingame",consoleplayer.nvtimeingame)
	local stopscroll = (
		not consoleplayer.spectator and
		not playerisdone and 
		nuvote.prev_gamestate == nuvote.GS_GAME and
		nv_state == NV_DEFAULT
	)
	if stopscroll then
		nuvote.timehint_offset = 0
		return
	end

	-- Don't scroll for the first 5 secs. Allow player to read before scrolling
	if consoleplayer.nvtimeingame < (TICRATE * 5) then return end

	-- scroll every half second
	if leveltime % (TICRATE / 2) != 0 then return end
	nuvote.timehint_offset = $+1

end)

function nuvote.printclockhelp(p)
	local helptxttbl = {
		"\151In console [backtick], type:",
		"\133nv_showtime \130true/false",
		"\133nv_ampm \130true/false",
		"\133nv_tzoffset \130-5",
		"\140Any \133red\140 alone will give details"
	}
	-- "\133Red\140=setting, \130yellow\140=example.", 
	chatprintf(p,table.concat(helptxttbl,"\128\n"))
end

COM_AddCommand("nvclockhelp",nuvote.printclockhelp)

addHook("PlayerMsg", function(source, msgtype, target, msg)
	if msgtype != 0 then return end
	if (msg != "!nvclockhelp") then return end
    nuvote.printclockhelp(source)
    return false
end)

local function drawNVDEFAULT(v,player)
	-- nuvote.drawVoteBg(v,player)
	nuvote.drawClock(v,player)
	drawDebug(v)
end

local enterrouletteframe = nil
local spinnerchoices = nil
local overshoot_amt = nil
local roulette_nexts = 0
local t_roulette_overshoot = 2 * TICRATE
local t_roulette_correct = t_roulette_overshoot + 10
local t_roulette_transition = t_roulette_correct + 10
local function drawSpinnerRungs(v,player,fx,fy,rungs,time)
	local textheight = 8
	local tmarginspace = 3
	local rungheight = textheight + tmarginspace
	local totalrungheight = (rungheight * #rungs)
	local targetspinnery = (rungheight * (#rungs-2))
	local inity = fy - totalrungheight
	local keyframes = {
		{x=fx,y=0,t=0,i='linear'},
		{x=fx,y=targetspinnery + overshoot_amt,t=t_roulette_overshoot,i='ease_out'},
		{x=fx,y=targetspinnery,t=t_roulette_correct,i='linear'},
	}
	local offsets = calcAnimPosition(keyframes,time,v)

	if offsets.y >= roulette_nexts then
		roulette_nexts = offsets.y/rungheight * rungheight + rungheight
		S_StartSound(nil,sfx_s1ba,player)
		-- v.drawFill(curx,cury,50,50,160)
	end

	-- v.drawString(cury,curx,offsets.y,0,"thin")
	for i,r in ipairs(rungs) do
		local ry = inity + ((i-1) * rungheight) + offsets.y
		if ry <= fy and ry >= (fy-22) then
			v.drawString(fx,ry,r,V_ALLOWLOWERCASE|V_SNAPTOBOTTOM,"center")
		end
	end
end

-- heavily inspired by Haya's combi ring. color var plucked straight from
local function drawSpinner(v,player,x,y)
	local color = min(leveltime & 31, 31 - (leveltime & 31)) + 160
	local sfx, sfy = x+128,y+22 -- 24 11
	v.drawFill(x, y, 247, 30, 10|V_SNAPTOBOTTOM) -- grey
	-- local rungs = {"a","I should see this result","a","a","a","a","a","a","a","a","a","a","a","a","a","a","a","a","a","a","a","a","a","a","a","a","This one should beshionwg","Smash Kart Circuit"}
	if enterrouletteframe != nil then
		local animtime = leveltime - enterrouletteframe
		drawSpinnerRungs(v,player,sfx,sfy,spinnerchoices,animtime)
	end
	v.drawFill(x, y, 247, 8, color|V_SNAPTOBOTTOM)
	v.drawFill(x, y+22, 247, 8, color|V_SNAPTOBOTTOM)
	v.drawString(x+85,y+2,"Decision in progress...",V_ALLOWLOWERCASE|V_50TRANS|V_SNAPTOBOTTOM,"small")

end

local function drawNVROULETTE(v,player)
	v.fadeScreen(0xFF00, 16)
	nuvote.drawVoteBg(v,player)
	local nvchoices = get_nv_choices()
	drawMapChoices(v,player,nvchoices)
	drawSpinner(v,player,29,94)

	if not wasmajorityforced then
		drawSpinner(v,player,29,94)
	end
	nuvote.drawClock(v,player)
	drawDebug(v)
end

local function drawNVRESULT(v,player)
	if server.nv_result == nil then 
		drawNVCHOICE(v,player)
		return
	end
	v.fadeScreen(0xFF00, 16)
	nuvote.drawVoteBg(v,player)
	local nvchoices = get_nv_choices()
	local resultMapId = nvchoices[server.nv_result][1]
	local reactionString = getReactionString(resultMapId)
	v.drawString(160,4, reactionString,V_ALLOWLOWERCASE,"center")
	-- draw the four choices
	drawMapChoices(v,player,nvchoices)
	nuvote.drawClock(v,player)
	drawDebug(v)
end

local function drawNVNEXT(v,player)
	v.fadeScreen(0xFF00, 16)
	v.drawString(160,4, "Up next...",V_ALLOWLOWERCASE,"center")
	local maps = {}
	if cq_hasmaps != nil then
		local cq_maps = cq_getallmaps()
		for _, map in ipairs(cq_maps) do
			table.insert(maps,map.id)
		end
	elseif aTourney !=nil then
		maps = aTourney.tourneyMapList
	end
	drawNextMap(v,96,18,75,maps[1],false)

	local hasaddl = false
	local remaining = min(#maps - 1,3)
	local midway = 160
	local miniwidth = 80
	local btwwidth = 10
	for i=2,min(#maps,4) do
		if i > #maps then continue end 
		local map = maps[i]
		local x = midway - (((miniwidth + btwwidth) / 2) * remaining) + ((i - 2) * (miniwidth + btwwidth))
		drawNextMap(v,x,129,50,map,false)
		hasaddl = true
	end
	if hasaddl then
		v.drawString(157,109, "On Deck:",V_ALLOWLOWERCASE,"center")
	end
	if #maps > 4 then
		v.drawString(292,150,"(+)")
	end
	nuvote.drawClock(v,player)
	drawDebug(v)
end

local animations = {
	votestart = {
		draw = drawVoteStart,
		timer = 0,
		totalticks = 3 * TICRATE
	},

	spinner = {
		draw = drawSpinnerRungs,
		timer = 0,
		totalticks = 3 * TICRATE
	}

}

local function playAnim(animName)
	nuvote.nva_off = false
	animations[animName].timer = 0
	nv_animation = animations[animName]
end


local function thinkNVDEFAULT()
	nuvote.checkVeto()
	nuvote.advanceMatchNum()
	checkForNewPlayers()
	pollMatchEnded()
end

local function thinkNVCHOICE()
	nuvote.nvchoicewarmup = $-1
	nuvote.checkToPlayMusic()
	nuvote.disableSneakerPanels()
	fetch_consoleplayer_values()
	searchPlayerInput()
	checkForNewPlayers()
	checkForSpec()
	countdownVote()
	countdownVoteCooldown()
	allowTeamChangePlaying()
	calcSuperMajorityReq()
	checkIfAllVoted()
end

local function thinkNVROULETTE()
	if wasmajorityforced then
		nv_state = NV_RESULT
		return
	end
	if enterrouletteframe == nil then return end
	local transition_time = enterrouletteframe + t_roulette_transition
	if transition_time < leveltime then
		nv_state = NV_RESULT
	end
end

local function thinkNVRESULT()
	countdownShowResult()
end

function nuvote.thinkNVWAIT()
	if nuvote.prev_gamestate != nuvote.GS_GAME then return end
	if leveltime < 3 * TICRATE then return end
	K_SetExitCountdown(0)
	G_ExitLevel()
end

local function thinkNVNEXT()
	K_SetExitCountdown(TICRATE*6000) -- Just keep setting this to ensure hardcode doesn't exit us 

	countdownShowResult()
end

local function onEnterNVDEFAULT()
	resetState()
	resetVoteTimers()
	resetChoices() -- there seems to be a delay in setting cvars
	nuvote.checkForExitlevel2()
end

local function onEnterNVCHOICE()
	-- print("onEnterNVCHOICE")
	updatePrngWithTimes()
	saveKartPositions()
	NV_playsound(nil,sfx_s3k68) -- final lap sound
	playAnim("votestart")
	loadMaps()
	loadNewChoices()
	resetVoteTimers()
	nuvote.disablePlayerControl()
	pacifyHMSoftlockDetection()
	checkExcludeLastList()
	calcSuperMajorityReq()
	resetPlayerInputs()
end

local function onEnterNVNEXT()
	updatePrngWithTimes()
	saveKartPositions()
	resetVoteTimers()
	nvshowresulttimer = 5 * TICRATE
	nuvote.disablePlayerControl()
	nuvote.playNVNEXTSounds()

end

local function onLeaveNVCHOICE()
	-- print("onLeaveNVCHOICE")
	if not calculatedresult then
		calculateResult()
	end
	nuvote.inputblackout = 50
end

local function onEnterNVROULETTE()
	-- nvdbp("entering nvroulette")
	enterrouletteframe = leveltime
	overshoot_amt = leveltime % 5
	local current_choices = get_nv_choices()
	local result_mapid = current_choices[min(server.nv_result,4)][1]
	local selection_order = {}
	local selection_order_ids = {}
	for p in players.iterate do
		if p.spectator then continue end
		local rungstr
		local mapid = current_choices[p.nv_choice][1]
		if p.nv_choice == 4 then
			rungstr = "Random"
		else
			rungstr = build_map_title(mapheaderinfo[mapid])
		end
		table.insert(selection_order, rungstr)
		table.insert(selection_order_ids, mapid)

	end
	local pselectoffset
	-- print("result mapid "..result_mapid)
	-- nvdbp(result_mapid,table.concat(selection_order_ids,","))
	for i,m in ipairs(selection_order_ids) do
		if m == result_mapid then
			pselectoffset = i
			break
		end
	end

	if pselectoffset == nil then
		print("Apparently anyone who voted previously is now gone. Choosing random.")
		nv_state = NV_RESULT
		return
	end

	spinnerchoices = {}
	for i=pselectoffset-1,pselectoffset-1+20 do
		local idx = (i % #selection_order)
		idx = ternary(idx==0,#selection_order,idx) -- stupid 1 indexing... 
		-- if you're reading this ping me on the best way to do this in lua lol
		table.insert(spinnerchoices, selection_order[idx])
	end

	nuvote.playVoteMusic("VOTEEA")
end

local function onEnterNVRESULT()
	playResultSounds()
	nuvote.playVoteMusic("VOTEEB")
	addResultToRecentMaps()
	NV_recordVotes()
	restoreHMSoftlockDetection()
end

local function onLeaveNVRESULT()
	-- clear roulette state
	-- nvdbp("Clearing roulette state")
	enterrouletteframe = nil
	spinnerchoices = nil
	overshoot_amt = nil
	roulette_nexts = 0
	nuvote.playVoteMusic("RACENT")
	-- print("onLeaveNVRESULT")
	-- enterLevel()
end

local function drawAnimation(v,player)
	nv_animation.draw(v,player,nv_animation.timer,nv_animation.totalticks)
	-- nv_animation.timer = $+1
	nuvote.drawClock(v,player)
	if nv_animation.timer > nv_animation.totalticks then
		nuvote.nva_off = true
	end
end


local stateFunctions = {
	[NV_DEFAULT] = drawNVDEFAULT,
	[NV_CHOICE] = drawNVCHOICE,
	[NV_ROULETTE] = drawNVROULETTE,
	[NV_RESULT] = drawNVRESULT,
	[NV_WAIT] = noop,
	[NV_NEXT] = drawNVNEXT,
}

local stateThinkFunctions = {
	[NV_DEFAULT] = thinkNVDEFAULT,
	[NV_CHOICE] = thinkNVCHOICE,
	[NV_ROULETTE] = thinkNVROULETTE,
	[NV_RESULT] = thinkNVRESULT,
	[NV_WAIT] = nuvote.thinkNVWAIT,
	[NV_NEXT] = thinkNVNEXT,
}

local onStateEnterHooks = {
	[NV_DEFAULT] = onEnterNVDEFAULT,
	[NV_CHOICE] = onEnterNVCHOICE,
	[NV_ROULETTE] = onEnterNVROULETTE,
	[NV_RESULT] = onEnterNVRESULT,
	[NV_WAIT] = noop,
	[NV_NEXT] = onEnterNVNEXT,
}

local onStateLeaveHooks = {
	[-1] = noop,
	[NV_DEFAULT] = noop,
	[NV_CHOICE] = onLeaveNVCHOICE,
	[NV_ROULETTE] = noop,
	[NV_RESULT] = onLeaveNVRESULT,
	[NV_WAIT] = noop,
	[NV_NEXT] = noop,
}

local function drawNuVote(v, player)
	-- nvdbp(nv_animation == nil,nv_animation.totalticks,nv_animation.timer)
	if not nv_enableNV.value then return end
	if (not nv_animation) or (nv_animation == nil) then
		stateFunctions[nv_state](v, player)
	else
		drawAnimation(v,player)
	end

	if player.nv_showhelp then
		drawNuVoteHelp(v, player)
	end
end

local function NuVoteThinker()
	if not nv_enableNV.value then return end
	-- state change detected
	if prevState != nv_state then
		-- nvdbp("just entered "..nv_state)
		onStateLeaveHooks[prevState]()
		onStateEnterHooks[nv_state]()
	end 
	prevState = nv_state

	-- turn off animation officially in a real thinker
	if nv_animation then
		nv_animation.timer = $+1
		if nuvote.nva_off then
			nv_animation = nil
		end
	end
	-- standard state thinkers
	stateThinkFunctions[nv_state]()

	nuvote.prev_gamestate = nuvote.GS_GAME
end

local function NuVoteClassicVoteThinker()
	-- this is called on the VoteThinker hook
	if not nv_enableNV.value then return end

	nuvote.prev_gamestate = nuvote.GS_VOTE

	-- if we don't have a result, it's likely for a good reason. don't proceed.
	if server.nv_result == nil then return end

	if nv_state != NV_WAIT then 
		error("reached vote thinker without going to NV_WAIT")
		return
	 end

	 -- detect if ATO or commandqueue are loaded. They will attempt to load their maps here
	 nuvote.mapwasqueued = $-1
	-- print(nuvote.mapwasqueued)
	if maps_are_queued() then
		if (cq_hasmaps != nil) and (nv_honor_preempted_votes.value) and (server.nv_result != nil) and (not nuvote.mapwarpcasted) then
			-- cq_hasmaps - this feature is only available for commandqueue
			local result = get_nv_choices()[server.nv_result]
			local mapstring, isencore = G_BuildMapName(result[1]), result[2]
			local cmd = "addq \"map "..mapstring..ternary(isencore," -encore\"", "\"")
			nuvote.mapwarpcasted = true
			COM_BufInsertText(server, cmd)
			print("maps were added to the queue during vote. But we'll honor the result of the vote by queuing it.")
		end
		return
	end
	if nuvote.mapwasqueued > 0 then return end
	enterLevel()
end

local function NuVoteIntermissionThinker()
	-- catch exceptional cases where NV proceedings are preempted
	-- usually by the mod softlock failsafe
	local prev_state = nuvote.prev_gamestate
	nuvote.prev_gamestate = nuvote.GS_INT
	if not nv_enableNV.value then return end

	local cur_state = nuvote.GS_INT
	if cur_state != prev_state then
		nuvote.intermission_countdown = CV_FindVar("inttime").value * TICRATE
	else
		-- nvdbp("Inttime: "..nuvote.intermission_countdown..CV_FindVar("inttime").value.." "..CV_FindVar("advancemap").string)
		-- nvdbp("cur_state","nuvote.prev_gamestate")
		-- nvdbp(cur_state,nuvote.prev_gamestate)
		nuvote.intermission_countdown = $-1
		local preempt_failsafe = nuvote.nuvote_preempted and not maps_are_queued()
		-- nvdbp("premptstats",nuvote.nuvote_preempted,maps_are_queued())
		-- nvdbp("failsafe", preempt_failsafe)
		local nonvotedround = (CV_FindVar("advancemap").string != "Vote") or preempt_failsafe
		if nuvote.intermission_countdown < 6  and nonvotedround then
			local nextmapinfo = nuvote.randommap_helper("randommap")
			local mapid, isencore = nextmapinfo[1], nextmapinfo[2]
			nvdbp("Doing cast map warp from intthinker...")
			nuvote.cast_mapwarp(mapid,isencore)
		end
	end


	-- mostly ok. Result is already calc'd
	if nv_state == NV_RESULT or nv_state == NV_ROULETTE then
		nv_state = NV_WAIT
		return
	end

	-- oof.. people could have been voting...
	if nv_state == NV_CHOICE then
		print("NuVote was preempted. tallying available votes...")
		calculateResult()
		local resultmapinfo = get_nv_choices()[server.nv_result]
		local resultmapid, resultmapencore = resultmapinfo[1], resultmapinfo[2]
		local mapstr = build_map_title(mapheaderinfo[resultmapid])
		if resultmapencore then
			mapstr = $.."(Encore)"
		end
		print("We're going to "..mapstr.." next...")
		nv_state = NV_WAIT
		return
	end

	-- if someone used exitlevel during the game
	if nv_state == NV_DEFAULT then
		print("NuVote never loaded this round. Selecting a random map...")
		nv_state = NV_WAIT
		nuvote.nuvote_preempted = true
		return
	end


end

local function resetNuVoteState()
	nv_state = NV_DEFAULT
	resetState()
end

hud.add(drawNuVote, "game")  
addHook("ThinkFrame",NuVoteThinker)
addHook("VoteThinker",NuVoteClassicVoteThinker)
addHook("IntermissionThinker",NuVoteIntermissionThinker)
addHook("MapLoad",resetNuVoteState)

local function one_match_or_print(p,query)
	local matches = query_map(query)

	if #matches > 1 then
		CONS_Printf(p,"Found multiple matches from your query. Try mapid?")
		CONS_Printf(p,"\130MAPID\128 | \130NAME")
		for i, match in ipairs(matches) do
			CONS_Printf(p,match.extid.." | "..match.name)
		end
		return nil
	elseif #matches == 0 then
		CONS_Printf(p,"Uh.. no maps matched your query, in both name and mapid. Check your spelling.")
		return nil
	end

	return matches[1]
end

local function get_full_config()
	local max= (36*26+26)+100
	local mapinfo = {}

	local idx = 1
    for i=0,max do
		local map = nil
		local mapname, gt, delisted, hell, reason = "<error>", "Unknown", false, false, ""
		local extid=G_BuildMapName(i)
		if mapheaderinfo[i] != nil then
			map = mapheaderinfo[i]

			if not (map.typeoflevel & (TOL_RACE | TOL_MATCH)) then
				-- delisteds[extid] = {extid=extid,reason="Generated from SOC"}
				delisted, reason = true, "Generated from SOC"
			end

			if Set.contains(delisted_maps,i) then
				-- delisteds[extid] = {extid=extid,reason=delist_reasons[i]}
				delisted, reason = true, delist_reasons[i]
			end

			-- Start as hell neutral - where we let the SOC decide if it's hell
			hell = -1
			if Set.contains(hellified_maps,i) then hell = 1 end
			if Set.contains(unhellified_maps,i) then hell = 0 end

			gt = "Unknown"
			if map.typeoflevel & TOL_RACE then
				gt = "race"
			elseif map.typeoflevel & TOL_MATCH then
				gt = "battle"
			end
			mapname = build_map_title(map)
		elseif map_cache[i] != nil then
			map = map_cache[i]
			mapname = map.mapname
			gt = map.gt
			delisted = map.delisted
			hell = map.hell
			reason = map.reason
		else
			continue
		end
		mapinfo[idx] = {
			rawid = i,
			extid = extid,
			name = mapname,
			gametype = gt,
			delisted = delisted,
			hell = hell,
			preferred = Set.contains(preferred_maps,i),
			reason = reason
		}
		idx =$+1
	end
	return mapinfo
end

local config_path = "NuVoteConfig.txt"
local mapschema = {}
mapschema["0.1"] = {
	rawid = 1,
	extid = 2,
	name = 3,
	gametype = 4,
	delisted= 5,
	hell = 6,
	preferred = 7,
	reason= 8
}


local function save_config()
	if not isserver return end

	local config = get_full_config()
	CONS_Printf(server,"saving NuVote config")
	local fd = io.open(config_path,'w+')
	local ct = 1

	fd:write("rawid,extid,name,gametype,delisted,hell,preferred,reason\n")

	for i,row in ipairs(config) do
		local output = ""
		local concattable = {}
		for key,value in pairs(row) do
			if type(value) == "boolean" then
				value = ternary(value,1,0)
			end
			concattable[mapschema["0.1"][key]] = value
			-- output = output..value..","
		end
		-- output = string.sub(output,1,-2)
		output = table.concat(concattable,",").."\n"
		fd:write(output)
		ct =$+1
		fd:flush()
	end
	fd:flush()
	fd:close()
	print("Saved NuVote map config. Wrote "..ct.." lines.")
end

local function _delist_map(p,mapid,reason,skippersist)
	delist_reasons[mapid] = reason
	Set.add(delisted_maps,mapid)
	checkExcludeLastList()
	if skippersist != true then save_config() end
end

local function _relist_map(p,mapid,skippersist)
	delist_reasons[mapid] = nil
	Set.remove(delisted_maps,mapid)
	if skippersist != true then save_config() end
end

local function _unhellify_map(p,mapid,skippersist)
	Set.remove(hellified_maps,mapid)
	Set.add(unhellified_maps,mapid)
	if skippersist != true then save_config() end
end

local function _hellify_map(p,mapid,skippersist)
	Set.remove(unhellified_maps,mapid)
	Set.add(hellified_maps,mapid)

	-- check if both hell and preferred
	if Set.contains(preferred_maps,mapid) then
		local wmsg = G_BuildMapName(mapid).." is set to be both a hell map and preferred map."
		nvwarn(p,wmsg)
	end
	if skippersist != true then save_config() end
end

-- Let the mapheaders (SOC files) decide
local function _hellneutral_map(p,mapid,skippersist)
	Set.remove(hellified_maps,mapid)
	Set.remove(unhellified_maps,mapid)
	if skippersist != true then save_config() end
end

local function _preferrify_map(p,mapid,skippersist)
	Set.add(preferred_maps,mapid)
	-- check if both hell and preferred
	if Set.contains(hellified_maps,mapid) then
		local wmsg = G_BuildMapName(mapid).." is set to be both a hell map and preferred map."
		nvwarn(p,wmsg)
	end
	if skippersist != true then save_config() end
end

local function _unpreferrify_map(p,mapid,skippersist)
	Set.remove(preferred_maps,mapid)
	if skippersist != true then save_config() end
end

local function delist_map(p,...)
	local query = {}
	local reason = {}
	local reason_reached = false

	-- parse map query and reason

	for i,arg in ipairs({...}) do
		if (arg == "-r") or (arg == "-reason") then
			reason_reached = true
			continue
		end
		if reason_reached then
			table.insert(reason,arg)
		else
			table.insert(query,arg)
		end
	end

	local map = one_match_or_print(p,query)
	if map == nil then return end

	reason = table.concat(reason," ")
	if (reason == nil) or (reason == "") then
		reason = "Unknown"
	end

	-- do the delisting
	_delist_map(p,map.id,reason)
	CONS_Printf(p,"\129"..map.name.."\130("..map.extid..")\128 is now \133DELISTED \128; It will no longer show up for votes.")
end

local function relist_map(p,...)
	local map = one_match_or_print(p,{...})
	if map == nil then return end

	_relist_map(p,map.id)
	CONS_Printf(p,"\129"..map.name.."\130("..map.extid..")\128is now \131NOT DELISTED \128; It is up for votes.")
end

local function unhellify_map(p,...)
	local map = one_match_or_print(p,{...})
	if map == nil then return end

	_unhellify_map(p,map.id)
	CONS_Printf(p,"\129"..map.name.."\130("..map.extid..")\128is now \131NOT HELL \128; It will appear for normal votes.")
end

local function hellify_map(p,...)
	local map = one_match_or_print(p,{...})
	if map == nil then return end

	_hellify_map(p,map.id)
	CONS_Printf(p,"\129"..map.name.."\130("..map.extid..")\128is now \133IN HELL \128; It will show up less frequently.")
end

local function hellneutral_map(p,...)
	local map = one_match_or_print(p,{...})
	if map == nil then return end

	_hellneutral_map(p,map.id)
	local headerhell = (mapheaderinfo[map.id].menuflags & LF2_HIDEINMENU) != 0
	local hellind = ternary(headerhell,"\133HELL\128","\131NOT HELL\128")
	CONS_Printf(p,"\129"..map.name.."\130("..map.extid..")\128is now \152HELL NEUTRAL\128; SOC files will decide if it's hell. In this case the map is "..hellind)
end

local function preferrify_map(p,...)
	local map = one_match_or_print(p,{...})
	if map == nil then return end

	_preferrify_map(p,map.id)
	CONS_Printf(p,"\129"..map.name.."\130("..map.extid..")\128is now \136PREFERRED\128; It will show up more frequently.")
end

local function unpreferrify_map(p,...)
	local map = one_match_or_print(p,{...})
	if map == nil then return end

	_unpreferrify_map(p,map.id)
	CONS_Printf(p,"\129"..map.name.."\130("..map.extid..")\128is now \130NOT PREFERRED \128; It will not show up for votes so frequently.")
end

local function show_mapstatus(p,...)
	local map = one_match_or_print(p,{...})
	if map == nil then return end
	local hell_ind = ""
	if map.hell then
		hell_ind = "\133 [HELL]\128"
	end
	
	if map.isdelisted then
		local delist_reason = map.delistreason
		if (delist_reason == nil) or (delist_reason == "") then delist_reason = "Unknown" end
		CONS_Printf(p,"\129"..map.name.."\130("..map.extid..hell_ind.."\130)\128 is \133DELISTED \128Reason: "..delist_reason)
		return
	end

	local preferred_str = ""
	local also_hell_warn = ""
	if map.ispreferred then
		preferred_str = "\128 and is a \136PREFERRED MAP"
		if map.hell then
			also_hell_warn = "\133 and may show up regardless of nv_hell_as_nonrandom_choice_pct."
		end
	end
	CONS_Printf(p,"\129"..map.name.."\130("..map.extid..hell_ind.."\130)\128 is \131NOT DELISTED"..preferred_str..also_hell_warn)

end

local function show_delisted_maps(p)
	show_maps_eligibility(p,false)
end

local function show_eligible_maps(p)
	show_maps_eligibility(p,true)
end

local function show_maps_eligibility(p,show_eligible)
	local cur_line = ""
	local max_cols = 3
	local cur_col = 1
	local zttl_cutoff = 10
	for i = 1, #mapheaderinfo do 
		local map = mapheaderinfo[i]
		if show_eligible != Set.contains(delisted_maps,i) then
			local map_name = build_map_title(map)
			local mapext = G_BuildMapName(i)
			cur_line = cur_line..mapext.." "..string.sub(map_name,1,zttl_cutoff).."... | "
			cur_col = $+1
			if cur_col > max_cols then 
				CONS_Printf(p,cur_line)
				cur_col =1
				cur_line = ""
			end
		end
	end
	CONS_Printf(p,cur_line)
end

local function config_map(p,mapid,delisted,ishell,ispreferred,reason,skippersist)
	mapid,delisted,ishell,ispreferred,skippersist = tonumber(mapid),tonumber(delisted),tonumber(ishell),tonumber(ispreferred),(skippersist=="1" or skippersist=="true" or skippersist == true)
	-- nvdbp("id",mapid,"delisted",delisted,"ishell",ishell,"ispreferred",ispreferred,"reason",reason,"skippersist",skippersist)
	local mid_type = type(mapid)
	if mid_type != "number" then
		if mapid == nil then mid_type = "nil" end
		nvwarn(p,"Incorrect type for mapid passed to config_map. Expecting number (was "..mid_type..mapid.." )")
		return
	end

	if delisted then
		_delist_map(p,mapid,reason,skippersist)
	end

	-- Keep map hell neutral if 1 or 0 not explicity set
	if ishell == 1 then
		_hellify_map(p,mapid,skippersist)
	elseif ishell == 0 then
		_unhellify_map(p,mapid,skippersist)
	end

	if ispreferred then
		_preferrify_map(p,mapid,skippersist)
	end
end

local nv_loading_config = false
local function load_config_complete(ct)
	print("Finished loading "..ct.." map configurations")
	nv_loading_config = false
end

local function load_config(p,custom_path)
	if p == nil then return end
	local path = config_path

	if custom_path != nil then path = custom_path end
	local fd = io.open(path, 'r')

	if fd == nil then
		nvwarn(p,"No config file found at "..path..". First time running NuVote?")
		return
	end
	
	nv_loading_config = true
	local first_line = true
	local ct = 0
	local commands = {}
	map_cache = {}
	for line in fd:lines() do
		if first_line then
			first_line = false
			continue
		end

		local cfgs = {}
		for config in string.gmatch(line, "[^,]+") do
		  table.insert(cfgs, config)
		end
		local mapid,delisted,ishell,ispreferred,reason = tonumber(cfgs[1]), tonumber(cfgs[5]), tonumber(cfgs[6]), tonumber(cfgs[7]), cfgs[8]
		if reason == nil then reason = "" end
		-- I don't know why, but maps get out of sync if we don't configure thru a command
		table.insert(commands,{p,"__nv_config_map "..mapid.." "..delisted.." "..ishell.." "..ispreferred.." \""..reason.."\" 1"})
		-- config_map(p,mapid,delisted,ishell,ispreferred,reason,true) -- while loading we skip persisting to disk
		ct =$+1
		map_cache[mapid] = {
			mapname=cfgs[3], -- human readable name
			gt=cfgs[4], -- gametype - battle/race
			delisted=delisted,
			hell=ishell,
			preferred=ispreferred,
			reason=reason
		}
	end
	funcQueue.add_command_batch(commands,{load_config_complete,{ct}})
	print("Loading NuVote map configs.")
end

local function load_config_helper(p,custom_path)
	-- We want the config to be loaded only from the server's config file
	-- so if someone is sending the command from a client exe, it'll be routed 
	-- so that the server is sending the command
	if isserver then 
		load_config(server,custom_path)
		return
	end

	if custom_path == nil then custom_path = "" end
	COM_BufInsertText(server,"nv_load_config "..custom_path)
end

local initial_config_load = false
addHook("MapLoad",function()
	if not isserver then return end
	if not initial_config_load then
		load_config(server)
		initial_config_load = true
	end
end)
-- cg_schedule_func_sec(COM_BufInsertText,{server,"nv_load_config"},5)
-- cg_schedule_func_sec(load_config,{server},5)
-- cg_schedule_func_sec(add_command_to_buffer,{server,"nv_load_config"},5)

function nuvote.moveCursorUp()
    -- cury = max(0,cury - 1)
    cury = cury - 1
end

function nuvote.moveCursorDown()
    cury = min(200,cury + 1)
end

function nuvote.moveCursorLeft()
    -- curx = max(0,curx - 1)
    curx = curx - 1
end

function nuvote.moveCursorRight()
    -- curx = min(320,curx + 1)
    curx = curx + 1
end

function nuvote.increaseVar(var)
	return function()
		nuvote[var] = $ + (FRACUNIT/100)
	end
end

function nuvote.decreaseVar(var)
	return function()
		nuvote[var] = $- (FRACUNIT/100)
	end
end

function nuvote.test_map_odds()
	loadMaps()
	local total_runs = 50000
	local freqs = {}
	local num_encore = 0
	for i=1,total_runs do
		local res = nuvote.randommap_helper("randommap")
		nuvote.match_num = $ + 1
		local mapid,encore = res[1], res[2]
		if freqs[mapid] == nil then
			freqs[mapid] = 1
		else
			freqs[mapid] = $+1
		end
		num_encore = $ + ternary(encore,1,0)
	end

	local fd = io.open("NuVoteMapChanceReport.txt","w+")
	fd:write(
		table.concat({"Mapid","ct","total","ishell","preferred\n"},",")
	)
	local maplist = get_full_config()
	for _,map in pairs(maplist) do
		if map.rawid == nil then continue end
		local temp = "%s,%d,%d,%s,%s"
		local rolls = freqs[map.rawid] or 0
		map.hell = mapIsHell(map.rawid)
		local msg = temp:format(map.extid, rolls, total_runs, tostring(map.hell), tostring(map.preferred))
		print(msg)
		fd:write(msg.."\n")
	end
	fd:flush()
	fd:close()
end

local function toggleState()

	-- nvdbp("encoremode",encoremode)
    if nv_state == NV_DEFAULT then
        nv_state = NV_CHOICE
    elseif nv_state == NV_CHOICE then
		nv_state = NV_ROULETTE
	elseif nv_state == NV_ROULETTE then
        nv_state = NV_DEFAULT
	elseif nv_state == NV_RESULT then
		nv_state = NV_WAIT
	elseif nv_state == NV_WAIT then
		nv_state = NV_DEFAULT
	elseif nv_state == NV_NEXT then
		nv_state = NV_DEFAULT
	end
end

-- local function reload()
--     CONS_Printf(server,"Reloading script")
--     COM_BufInsertText(server,"addfile luamaking/nuvote.lua")
-- end

COM_AddCommand("mcl",nuvote.moveCursorLeft)
COM_AddCommand("mcr",nuvote.moveCursorRight)
COM_AddCommand("mcu",nuvote.moveCursorUp)
COM_AddCommand("mcd",nuvote.moveCursorDown)

-- COM_AddCommand("up1",nuvote.increaseVar("cur1"))
-- COM_AddCommand("dn1",nuvote.decreaseVar("cur1"))


-- COM_AddCommand("up2",nuvote.increaseVar("cur2"))
-- COM_AddCommand("dn2",nuvote.decreaseVar("cur2"))


-- COM_AddCommand("up3",nuvote.increaseVar("cur3"))
-- COM_AddCommand("dn3",nuvote.decreaseVar("cur3"))


-- COM_AddCommand("up4",nuvote.increaseVar("cur4"))
-- COM_AddCommand("dn4",nuvote.decreaseVar("cur4"))


-- COM_AddCommand("up5",nuvote.increaseVar("cur5"))
-- COM_AddCommand("dn5",nuvote.decreaseVar("cur5"))


-- COM_AddCommand("up6",nuvote.increaseVar("cur6"))
-- COM_AddCommand("dn6",nuvote.decreaseVar("cur6"))


-- COM_AddCommand("up7",nuvote.increaseVar("cur7"))
-- COM_AddCommand("dn7",nuvote.decreaseVar("cur7"))


-- COM_AddCommand("up8",nuvote.increaseVar("cur8"))
-- COM_AddCommand("dn8",nuvote.decreaseVar("cur8"))


-- COM_AddCommand("up9",nuvote.increaseVar("cur9"))
-- COM_AddCommand("dn9",nuvote.decreaseVar("cur9"))



	
COM_AddCommand("__nvoddsreport",nuvote.test_map_odds,COM_ADMIN)

COM_AddCommand("nvtest",function(user,param)
	if not CV_FindVar("nv_debug").value then return end
	if param == nil then
		param = "nil"
	end
	toggleState()
	-- for i=0,33 do
	-- 	nvdbp((i+1).." bits: ", 1 << i)
	-- end

	-- nvdbp(G_BuildMapName(0):sub(4,5))
	-- print(gamemap)
	-- test_NV_Random()
	-- print(testvar.value)
	-- nuvote.test_randommap_helper()
	end,COM_ADMIN)
	
-- COM_AddCommand("rl",reload)

-- -- COM_AddCommand("exitlevel",toggleState)a
COM_AddCommand("vote1",function(p) castVote(p,1) end)
COM_AddCommand("vote2",function(p) castVote(p,2) end)
COM_AddCommand("vote3",function(p) castVote(p,3) end)
COM_AddCommand("vote4",function(p) castVote(p,4) end)

COM_AddCommand("nv_delist",delist_map,COM_ADMIN)
COM_AddCommand("nv_relist",relist_map,COM_ADMIN)
COM_AddCommand("nv_unhellify",unhellify_map,COM_ADMIN)
COM_AddCommand("nv_hellify",hellify_map,COM_ADMIN)
COM_AddCommand("nv_hellneutralize",hellneutral_map,COM_ADMIN)
COM_AddCommand("nv_prefer",preferrify_map,COM_ADMIN)
COM_AddCommand("nv_unprefer",unpreferrify_map,COM_ADMIN)
COM_AddCommand("nv_show_delisted",show_delisted_maps)
COM_AddCommand("nv_show_mapstatus",show_mapstatus)
COM_AddCommand("nv_save_config",save_config,COM_ADMIN)
COM_AddCommand("nv_load_config",load_config_helper,COM_ADMIN)
COM_AddCommand("__nv_config_map",config_map,COM_ADMIN)

COM_AddCommand("randommap",nuvote.randommap,COM_ADMIN)

local nv_help_entries = {
	nv_enableNV = "\139{On|Off}\128 Toggles NuVote functionality",
	__nvoddsreport = "Creates a report for liklihood for all maps. \130Warning: \128Do not run during an active netgame.",
	__nv_config_map = "Internal command; do not call",
	vote1 = "Sends a vote for the top left choice.",
	vote2 = "Sends a vote for the top right choice.",
	vote3 = "Sends a vote for the bottom left choice.",
	vote4 = "Sends a vote for the bottom right (random) choice.",
	nv_showtime = "\139{On|Off}\128 Toggles whether to display the clock on the bottom left",
	nv_ampm = "\139{On|Off}\128 Toggles 12hr vs 24hr clock.",
	nv_tzoffset = "\139{mm:hh}\128 Offset from UTC for the clock. Type the command by itself for more info. Check https://greenwichmeantime.com/articles/clocks/global/ for your offset",
	nv_donottrackme = "\139{On|Off}\128 Strikes your username and color from any popularity tracking.",
	nv_showsmr = "\139{On|Off}\128 Toggles whether users can see how many votes are required for majority rule to take effect",
	nv_honor_preempted_votes = "\139{On|Off}\128 Applies only when command queue or ATO is present. In cases where a map is queued while a vote is taking place, toggles whether result of vote is queued also.",
	nv_include_total_state_in_tracking = "\139{On|Off}\128 In popularity tracking, the state for which maps are enabled, preferred, and in hell is encoded. It's a lot. Toggle this to exlcude.",
	nv_supermajoritypct = "\139{51:100}\128 Percent of players required to vote for a map for majority rule to take effect",
	nv_superminoritypct = "\139{0:50}\128 Minimum percent requirement for a choice to be eligible for the roulette. If none make it, the choice with most votes wins.",
	nv_debug = "\139{On|Off}\128 Toggles NuVote debug info. \130WARNING:\128LOTS of HUD clutter. Not recommended.",
	nv_random_induces_hell = "\139{On|Off}\128 Toggles rule where if everyone votes random, a hell map will be chosen instead.",
	nv_hell_as_nonrandom_choice_pct = "\139{0:100}\128 Percent chance a hell map will appear as non-random choice. Hell choices are indicated as hell.",
	nv_hell_as_random_choice_pct = "\139{0:100}\128 Percent chance a hell map will be the random choice (triggers \"This is bad\" clip).",
	nv_hell_as_randommap_pct = "\139{0:100}\128 Percent chance of a hell map being loaded when advancemap=Random, or if randommap is called",
	nv_encore_round_pct = "\139{0:100}\128 Percent chance an encore round ocurrs. Encore rounds will have 3 choices as encore, and 1 normal.",
	nv_encore_choice_pct = "\139{0:100}\128 Percent chance any choice will be encore",
	nv_exclude_last_n = "\139{0:MAXINT}\128 Number of recently played maps to exclude from the choice algorithm",
	exitlevel2 = "Acts similar to exitlevel, except it triggers the NuVote voting screen too. exitlevel is still available.",
	randommap = "\133Overwrites hostmod's version.\128 Loads a random map.",
	nv_delist = "\139QUERY [ -reason|-r REASON ...]\128 Queries a map by name or id. Result map will be \133DELISTED\128 - it will no longer show up for voting.",
	nv_relist = "\139QUERY\128 Queries a map by name or id. Result map will be will be removed from the list of \133DELISTED\128 maps; It will show up for voting.",
	nv_unhellify = "\139QUERY\128 Queries a map by name or id. Result map will be removed from the list of hell maps.",
	nv_hellify = "\139QUERY\128 Queries a map by name or id. Result map will be moved to the list of hell maps.",
	nv_hellneutralize = "\139QUERY\128 Queries a map by name or id. Result map's hell status will only be determined by the present .soc files loaded.",
	nv_prefer = "\139QUERY\128 Queries a map by name or id. Result map will be added as a \136PREFERRED\128 map. \136PREFERRED\128 maps will show up more frequently.",
	nv_unprefer = "\139QUERY\128 Queries a map by name or id. Result map will be removed from the \136PREFERRED\128 maps list.",
	nv_preferred_every = "\139{0:99999}\128 How often the preferred map pool will be pulled from. 20 means every 20 rounds, 0 means never.",
	nv_preferred_ct = "\139{0:4}\128 When NuVote decides to pull from the preferred map pool, how many should it take? 0 means none. 4 means all.",
	nv_show_mapstatus = "\139QUERY\128 Queries a map by name or id. Prints the status of the result map. Whether it's \133DELISTED\128, in hell, and/or \136PREFERRED\128.",
	nv_show_delisted = "\133Deprecated\128 Prints a list of all maps which will not show up for voting.",
	nv_save_config = "Saves current map settings for each loaded map to disk. This is done every time a map setting is modified, btw.",
	nv_load_config = "\139[FILEPATH]\128 Loads current map settings from disk, optionally from file specified at FILEPATH.",
	nv_show_vote_bg = "\139{On|Off}\128 Toggles whether to display the background seen in the vanilla voting screen.",
	nv_play_vote_music = "\139{On|Off}\128 Toggles whether to play the music heard in the vanilla voting screen.",
	nvclockhelp = "Shows available commands to control the clock on the bottom left.",
	nv_showhelp = "Displays NuVote help page.",
	nv_hidehelp = "Hides NuVote help page.",
	nv_help = "\139NVCOMMAND\128 Prints help text for the provided NuVote command."
}

COM_AddCommand("nv_help",function(p,pcmd)
	local cmds = {}
	local matched_cmds = {}
	for cmd,helptxt in pairs(nv_help_entries) do
		if pcmd == cmd then
			CONS_Printf(p,"\143"..cmd.."\128 "..helptxt)
			return
		end
		table.insert(cmds,cmd)
	end
	local msg = ""
	if pcmd != nil then
		msg = "Query "..pcmd.." didn't match any available commands. Could be mispelled? Above are the available NuVote commands:"
	else
		msg = "\131Welcome to NuVote! \128Type \130nv_showhelp\128 for a basic gist. Type \130nv_help COMMAND\128 for breif info on a COMMAND above (more available with \137PGUP\128). Check out the readme.md or MB for full documentation."
	end
	CONS_Printf(p,"")
	local cmds_str = "\130"..table.concat(cmds,"\128 \n \130").."\128"
	CONS_Printf(p,cmds_str)
	CONS_Printf(p,msg)
end)


addHook("NetVars",function(net)
	nvvotetimer = net($)
	nvshowresulttimer = net($)
	nv_state = net($)
	-- nv_choices = net($)
	
	nv_animation = net($)
	nv_recent_maps = net($)

	randominducedhell = net($)
	wasmajorityforced = net($) 
	calculatedresult = net($)

	delisted_maps = net($)
	delist_reasons = net($)
	preferred_maps = net($)
	hellified_maps = net($)
	unhellified_maps = net($)
end)

	-- print("\120120 \121121 \122122 \123123 \124124 \125125 \126126 \127127 \128128 \129129 \130130 \131131 \132132 \133133 \134134 \135135 \136136 \137137 \138138 \139139 \140140 \141141 \142142 \143143 \144144 \145145 \146146 \147147 \148148 \149149 \150150 \151151 \152152")
