print("Loading NuVote Help....")

-- Configurable offsets for ui elements
local nvge = {
    nvw = {x=43,y=22}, -- help window top left location
    basepnl={x=43,y=22},
    basepnlbr={x=253,y=168},
    titletxt={x=60,y=2},
    titlegfx={x=120,y=35},
    subttl={x=10,y=26},
    helptxt1={x=13,y=46},
    samplev={x=179,y=44},
    smtxlh={x=7,y=0},
    helpgfx1={x=0,y=0},
    ccardttl={x=30,y=86},
    ccardtbl={x=30,y=99},
    bugrpt={x=30,y=319},
    cgsignoff={x=0,y=272},
    dismiss={x=70,y=162}
}

local nvguiswaporder = {
    "nvw",
    "basepnl",
    "basepnlbr",
    "titletxt",
    "titlegfx",
    "subttl",
    "helptxt1",
    "samplev",
    "smtxlh",
    "helpgfx1",
    "ccardttl",
    "ccardtbl",
    "bugrpt",
    "cgsignoff",
    "dismiss"
}

local cge = {
    tl = {x=-3,y=-5},
    tr = {x=33,y=-5},
    bl = {x=-10,y=19},
    br = {x=27,y=19},
    dim = {x=25,y=15},
    no = {x=10,y=3},
}

local cgeorder = {
    "tl",
    "tr",
    "bl",
    "br",
    "dim",
    "no",
}

local itbl = {
    bgtl = {x=0,y=0},
    bgbr = {x=176,y=218},
    txt = {x=3,y=3},
    offy = {x=10,y=8},
    offx = {x=144,y=0},
    cll = {x=142,y=16},
    cll2 = {x=28,y=16}
}

local itblorder = {
    "bgtl",
    "bgbr",
    "txt",
    "offy",
    "offx",
    "cll",
    "cll2"
}

local nvworder = {
    "tl",
    "wh",
    "ttlh",
    "maxscl",
    "minscl",
    "off"
}

local NVWindow = {
    tl = {x=0,y=0},
    wh = {x=253,y=168}, -- help window width and height
    ttlh = {x=12,y=12}, -- header height, footer height
    maxscl = {x=-188,y=-173},
    minscl = {x=0,y=-12},
    off = {x=0,y=0}
}

function NVWindow.drawString(nvw,v,p,x,y,str,flags,align)
    -- nvdbp("ds",x,nvw.tl.x)
    if (y - nvge.nvw.y) < 0 then return end
    if (y - nvge.nvw.y) > nvw.wh.y - nvw.ttlh.y then return end
    local cx, cy = (x + nvw.tl.x), (y + nvw.tl.y)
    if not align then align = "left" end
    v.drawString(cx,cy,str,flags,align)
end

function NVWindow.drawPatchScaled(nvw,v,p,x,y,scale,patch,flags)
    -- nvdbp("ds",x,nvw.tl.x)
    if (y - (nvge.nvw.y * FRACUNIT)) < 0 then return end
    if (y - (nvge.nvw.y * FRACUNIT)) > (nvw.wh.y * FRACUNIT) - (nvw.ttlh.y * FRACUNIT) then return end
    local cx, cy = (x + (nvw.tl.x * FRACUNIT)), (y + (nvw.tl.y * FRACUNIT))
    v.drawScaled(cx,cy,scale,patch,flags)
end

function NVWindow.drawFill(nvw,v,p,x,y,w,h,color)
    -- Draw only up to bounds
    -- y = y=nvge.nvw.y + nvge.ccardtbl.y + offset [maxscl, minscl] +  +itbl.bgtl.y
    local disttobottom = (nvge.nvw.y + nvw.wh.y - nvw.ttlh.y) - y
    local distfromtop = (y+h) - (nvge.nvw.y+nvw.ttlh.x)
    local maxheight = nvw.wh.y - (nvw.ttlh.y + nvw.ttlh.x)
    h = min(h, disttobottom)
    h = min(h, distfromtop)
    h = min(h, maxheight)
    y = max(y, nvge.nvw.y + nvw.ttlh.x)
    y = min(y, nvge.nvw.y + nvw.wh.y - nvw.ttlh.y)
    
    v.drawFill(x,y,w,h,color)
end

function NVWindow.draw(nvw,v,p,x,y)

    -- black border
    local bx, by, bh, bw = x + nvw.tl.x-1, y + nvw.tl.y-1,nvw.wh.x+2,nvw.wh.y+2
    v.drawFill(bx,by,bh,bw,31)

    -- base sheet
    local fx, fy, fh, fw = x + nvw.tl.x, y + nvw.tl.y,nvw.wh.x,nvw.wh.y
    v.drawFill(fx,fy,fh,fw,0)

    -- acutal elements
    for i, df in ipairs(nvw.drawFunctions) do
        local offset = p.nv_showhelpoffset
        df.func(v,p,df.x + x,df.y + y+offset)
    end
    
    -- title bar
    local ttx, tty, ttw, tth = x + nvw.tl.x, y + nvw.tl.y,nvw.wh.x,nvw.ttlh.x
    v.drawFill(ttx,tty,ttw,tth,4)
    v.drawString(x+nvge.titletxt.x,y+nvge.titletxt.y,"Welcome to NuVote!!",textflags)

    -- footer bar
    local tbx, tby, tbw, tbh = x + nvw.tl.x, y + nvw.wh.y - nvw.ttlh.y, nvw.wh.x,nvw.ttlh.x
    v.drawFill(tbx,tby,tbw,tbh,4)

    if consoleplayer and consoleplayer.spectator then
        local msg = "Press \131ACCEL\128/\133BRAKE\128 to scroll, \130ITEM\128/\130DRIFT\128 to dismiss"
        v.drawString(x+nvge.dismiss.x-23, y+nvge.dismiss.y,msg,textflags,"small")
    else
        v.drawString(x+nvge.dismiss.x, y+nvge.dismiss.y,"Press any button to dismiss",textflags,"small")
    end
end

-- controllers for configuring element placement
local allptr = 1
local guiptr = 1
local cgeptr = 1
local itblptr = 1
local wptr = 1

local allgui = {
    {
        data=nvge,
        order=nvguiswaporder,
        ptr=guiptr,
    },
    {
        data=cge,
        order=cgeorder,
        ptr=cgeptr
    },
    {
        data=itbl,
        order=itblorder,
        ptr=itblptr
    },
    {
        data=NVWindow,
        order=nvworder,
        ptr=wptr
    }
}

local allguiorder = {
    "nvge",
    "cge",
    "itbl",
    "nvw"
}

local function increaseAllGuiPtr()
    allptr = (allptr % #allguiorder) + 1
end

local function decreaseAllGuiPtr()
    local guio = allgui[allptr]
    allptr = $-1
    if allptr == 0 then
        allptr = #allguiorder
    end
end

local function increaseGuiPtr()
    local guio = allgui[allptr]
    guio.ptr = (guio.ptr % #guio.order) + 1
end

local function decreaseGuiPtr()
    local guio = allgui[allptr]
    guio.ptr = $-1
    if guio.ptr == 0 then
        guio.ptr = #guio.order
    end
end

local function moveCursorUp()
    local guio = allgui[allptr]
    guio.data[guio.order[guio.ptr]].y = $ - 1
end

local function moveCursorDown()
    local guio = allgui[allptr]
    guio.data[guio.order[guio.ptr]].y = $ + 1
end

local function moveCursorLeft()
    local guio = allgui[allptr]
    guio.data[guio.order[guio.ptr]].x = $ - 1
end

local function moveCursorRight()
    local guio = allgui[allptr]
    guio.data[guio.order[guio.ptr]].x = $ + 1
end



-- COM_AddCommand("apr",increaseAllGuiPtr)
-- COM_AddCommand("apl",decreaseAllGuiPtr)
-- COM_AddCommand("mpl",decreaseGuiPtr)
-- COM_AddCommand("mpr",increaseGuiPtr)
-- COM_AddCommand("mcu",moveCursorUp)
-- COM_AddCommand("mcd",moveCursorDown)
-- COM_AddCommand("mcl",moveCursorLeft)
-- COM_AddCommand("mcr",moveCursorRight)

local colors = {
	[0] = 0,
	[1] = 215
}

local function enable_nvhelp(p)
    p.nv_showhelp = true
    p.nv_showhelpoffset = NVWindow.minscl.y
end

COM_AddCommand("nv_showhelp",function(p)
    enable_nvhelp(p)
end)

addHook("PlayerMsg", function(source, msgtype, target, msg)
	if msgtype != 0 then return end
	if (msg:sub(1, 7):lower() != "!nvhelp") then return end
	if #msg != 7 then return end
    nvdbp("enabled help")
    enable_nvhelp(source)
    return true
end)

local function playerAccelForced(p)
    return p.kartstuff[k_sneakertimer] > 0
end

local function disableHelp(p,src)
    if src == "accel" and playerAccelForced(p) then
        return 
    end
    p.nv_showhelp = false
end

-- ugh.. I am duplicating some controller handling code here.
-- The only reason I'm doing this is that it's cost me SO MUCH 
-- time to get it working. 
-- TBH, I don't want to potentially spend another few months
-- getting it fixed, if I refactor it incorrectly somehow.

local localprevinput = {
	accel = false,
    drift = false,
    item = false,
    brake = false
}

addHook("MapLoad",function()
    localprevinput = {
        accel = false,
        drift = false,
        item = false,
        brake = false
    }
	for p in players.iterate do
		p.nv_showhelp = false
	end
end)

local scrollamt = 3
local function handleinputs()
    if consoleplayer == nil then return end
    if not consoleplayer.nv_showhelp then return end
    local cmd = consoleplayer.cmd
    local cmds = {}
    cmds["accel"] = (cmd.buttons & BT_ACCELERATE) > 0
    cmds["drift"] = (cmd.buttons & BT_DRIFT) > 0
    cmds["brake"] = (cmd.buttons & BT_BRAKE) > 0
    cmds["item"] = (cmd.buttons & BT_ATTACK) > 0
	cmds["up"] = (cmd.buttons & BT_UP) != 0
	cmds["down"] = (cmd.buttons & BT_DOWN) != 0

    local scroll_btns = ternary(
        consoleplayer.spectator,
        {"accel","brake","up","down"},
        {"up","down"}
    )

	for ckey,cval in pairs(cmds) do
        if ckey:inTable(scroll_btns) then continue end
		if cval and cval != localprevinput[ckey] then
			disableHelp(consoleplayer,ckey)
		end
		localprevinput[ckey] = cval
	end

    local nvw = NVWindow
    if cmds["up"] or cmds["brake"] then
        -- nvdbp(cmds["up"],consoleplayer.nv_showhelpoffset,nvw.minscl.x,consoleplayer.nv_showhelpoffset + scrollamt)
        consoleplayer.nv_showhelpoffset = min(nvw.minscl.x, $ + scrollamt)
    elseif cmds["down"] or cmds["accel"] then
        consoleplayer.nv_showhelpoffset = max(nvw.maxscl.x, $ - scrollamt)

    --If beyond either scroll target, creep towards it
    elseif consoleplayer.nv_showhelpoffset > nvw.minscl.y then
        consoleplayer.nv_showhelpoffset = $- 1
    elseif consoleplayer.nv_showhelpoffset < nvw.maxscl.y then
        consoleplayer.nv_showhelpoffset = $ + 1
    end
end
addHook("ThinkFrame", handleinputs)

COM_AddCommand("nv_hidehelp",function(p)
    disableHelp(p,"console")
end)

local textflags = V_ALLOWLOWERCASE


local function drawConfigCard(v,p,x,y)
    
    local hellforced = ternary(CV_FindVar("nv_random_induces_hell").value,"\133is guaranteed.", "\152is not guaranteed.")
    local tbl = {
        {"cvar name", "Description", "Value"},
        {CV_FindVar("nv_supermajoritypct").name, "Votes for majority rule", CV_FindVar("nv_supermajoritypct").value.."%"},
        {CV_FindVar("nv_superminoritypct").name, "Votes required for roulette", CV_FindVar("nv_superminoritypct").value.."%"},
        {CV_FindVar("nv_hell_as_random_choice_pct").name, "Chance for hell in random", CV_FindVar("nv_hell_as_random_choice_pct").value.."%"},
        {CV_FindVar("nv_hell_as_randommap_pct").name, "Chance for hell with randommap", CV_FindVar("nv_hell_as_randommap_pct").value.."%"},
        {CV_FindVar("nv_hell_as_nonrandom_choice_pct").name, "Chance for hell in open", CV_FindVar("nv_hell_as_nonrandom_choice_pct").value.."%"},
        {CV_FindVar("nv_encore_choice_pct").name, "Chance for encore", CV_FindVar("nv_encore_choice_pct").value.."%"},
        {"<DRAW HELLICON>", "Indicates if map is hell", ""},
        {CV_FindVar("nv_random_induces_hell").name, "If all players vote random, hell "..hellforced,""},
        {"nv_showtime", "Clock toggle (local)",tostring(nvs_get(consoleplayer.nvsettings,"nv_showtime"))},
        {"nv_ampm", "12hr/24hr clock toggle (local)",tostring(nvs_get(consoleplayer.nvsettings,"nv_ampm"))},
        {"nv_tzoffset", "Clock UTC offset (local)",nvs_get(consoleplayer.nvsettings,"nv_tzoffset")},

    }

    -- background for table
    local cury = itbl.txt.y
    NVWindow.drawFill(NVWindow,v,p,x+itbl.bgtl.x,y+itbl.bgtl.y,itbl.bgbr.x,itbl.bgbr.y,20)

    -- individual cells
    for i, row in pairs(tbl) do
        NVWindow.drawFill(NVWindow,v,p,x+itbl.txt.x-1,y+cury-1,itbl.cll.x,itbl.cll.y,12)
        NVWindow.drawFill(NVWindow,v,p,x+itbl.txt.x+itbl.offx.x-1,y+cury-1,itbl.cll2.x,itbl.cll2.y,12)
        NVWindow.drawString(NVWindow,v,p,x+itbl.txt.x,y+cury,row[2],textflags,"thin")
        NVWindow.drawString(NVWindow,v,p,x+itbl.txt.x+itbl.offx.x,y+cury,row[3],textflags,"thin")
        if row[1] != "<DRAW HELLICON>" then
            NVWindow.drawString(NVWindow,v,p,x+itbl.txt.x,y+cury+itbl.offy.x,"("..row[1]..")",textflags,"small")
        else
            local patchName = "HELLSPB"
            -- local patchName = "KEGGRANK"
            if v.patchExists(patchName) then
                local patch = v.cachePatch(patchName)
                NVWindow.drawPatchScaled(
                    NVWindow, v, p,
                    (x+itbl.txt.x+itbl.offx.x) * FRACUNIT,
                    (y+cury) * FRACUNIT,
                    75 * FRACUNIT / 100,
                    patch,
                    0
                )
            else
                NVWindow.drawString(NVWindow,v,p,x+itbl.txt.x+itbl.offx.x,y+cury,"HELL",textflags,"small")
            end
            -- NVWindow.drawPatch(....)
        end
        
        cury = $ + itbl.offy.y + itbl.offy.x
    end
end

local function drawSampleVote(v,p,x,y)
    NVWindow.drawFill(NVWindow,v,p,x+cge.tl.x,y+cge.tl.y,cge.dim.x,cge.dim.y,15)
    NVWindow.drawFill(NVWindow,v,p,x+cge.tr.x,y+cge.tr.y,cge.dim.x,cge.dim.y,15)
    NVWindow.drawFill(NVWindow,v,p,x+cge.bl.x,y+cge.bl.y,cge.dim.x,cge.dim.y,15)
    NVWindow.drawFill(NVWindow,v,p,x+cge.br.x,y+cge.br.y,cge.dim.x,cge.dim.y,15)
    
    NVWindow.drawString(NVWindow,v,p,x+cge.tl.x+cge.no.x,y+cge.tl.y+cge.no.y,"1",textflags,"thin")
    NVWindow.drawString(NVWindow,v,p,x+cge.tr.x+cge.no.x,y+cge.tr.y+cge.no.y,"2",textflags,"thin")
    NVWindow.drawString(NVWindow,v,p,x+cge.bl.x+cge.no.x,y+cge.bl.y+cge.no.y,"3",textflags,"thin")
    NVWindow.drawString(NVWindow,v,p,x+cge.br.x+cge.no.x,y+cge.br.y+cge.no.y,"4",textflags,"thin")
end

local function drawVotePrimer(v,p,x,y)
    -- First free text part
    NVWindow.drawString(NVWindow,v,p,x + nvge.subttl.x, y + nvge.subttl.y,"The alternative voting system where majority rules!", textflags,"thin")
    local lineheight = nvge.smtxlh.x
    local ctrltxt = {
        "* \130Aim/Steer\128 to move cursor",
        "* \131ACCEL\128 to select",
        "* \150SPECTATE\128 to unselect",
        "    \152OR\128",
        "* \140Chat\128 the number corresp. to your pref."
        }
    for i, txt in ipairs(ctrltxt) do
        NVWindow.drawString(NVWindow,v,p,x+nvge.helptxt1.x, y + nvge.helptxt1.y + ((i-1) * lineheight),txt, textflags,"small")
    end
    NVWindow.drawString(NVWindow,v,p,x+ nvge.ccardttl.x,y + nvge.ccardttl.y,"This server's \139CONFIG CARD",textflags)

    -- footer after the table cells
    NVWindow.drawString(NVWindow,v,p,x+nvge.bugrpt.x,y+nvge.bugrpt.y,"For any issues, please contact cglitcher.", textflags,"small")
end

NVWindow.drawFunctions = {drawSampleVote}

local function drawNuVoteHelp(v,p)
    NVWindow.drawFunctions = {
        {func=drawVotePrimer,x=0,y=0},
        {func=drawSampleVote,x=nvge.samplev.x,y=nvge.samplev.y},
        {func=drawConfigCard,x=nvge.ccardtbl.x,y=nvge.ccardtbl.y}
    }
    NVWindow.draw(NVWindow,v,p,nvge.nvw.x,nvge.nvw.y)

    -- local aelementname = allguiorder[allptr]
    -- local guio = allgui[allptr]
    -- local elementname = guio.order[guio.ptr]
    -- local curxy = guio.data[elementname]
    -- local curx, cury = curxy.x, curxy.y
    -- v.drawFill(curx,cury,1,1,colors[leveltime / 8 % 2])
	-- v.drawString(230,160,"Categry: "..aelementname,0 ,"small")
	-- v.drawString(230,170,"element: "..elementname,0 ,"small")
    -- v.drawNum(280,180,curx)
    -- v.drawNum(310,180,cury)
end

rawset(_G,"drawNuVoteHelp", drawNuVoteHelp)
